#
# This file is part of pySMT.
#
#   Copyright 2015 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from math import log, ceil

from pysmt.exceptions import SolverAPINotFound

try:
    import pyboolector
except ImportError:
    raise SolverAPINotFound


from pysmt.solvers.solver import (
    IncrementalTrackingSolver,
    UnsatCoreSolver,
    Converter,
    SolverOptions,
)
from pysmt.solvers.smtlib import SmtLibBasicSolver, SmtLibIgnoreMixin
from pysmt.solvers.eager import EagerModel
from pysmt.walkers import DagWalker
from pysmt.exceptions import (
    SolverReturnedUnknownResultError,
    ConvertExpressionError,
    PysmtValueError,
)
from pysmt.decorators import clear_pending_pop, catch_conversion_error
from pysmt.logics import QF_BV, QF_UFBV, QF_ABV, QF_AUFBV, QF_AX
from pysmt.constants import to_python_integer


class BoolectorOptions(SolverOptions):
    def __init__(self, **base_options):
        SolverOptions.__init__(self, **base_options)
        if self.random_seed is not None:
            raise PysmtValueError("BTOR Does not support Random Seed setting.")

        # Disabling Incremental usage is not allowed.
        # This needs to be set to 1
        self.incrementality = True
        self.internal_options = [
            pyboolector.BTOR_OPT_MODEL_GEN,
            pyboolector.BTOR_OPT_INCREMENTAL,
            pyboolector.BTOR_OPT_INCREMENTAL_SMT1,
            pyboolector.BTOR_OPT_INPUT_FORMAT,
            pyboolector.BTOR_OPT_OUTPUT_NUMBER_FORMAT,
            pyboolector.BTOR_OPT_OUTPUT_FORMAT,
            pyboolector.BTOR_OPT_ENGINE,
            pyboolector.BTOR_OPT_SAT_ENGINE,
            pyboolector.BTOR_OPT_AUTO_CLEANUP,
            pyboolector.BTOR_OPT_PRETTY_PRINT,
            pyboolector.BTOR_OPT_EXIT_CODES,
            pyboolector.BTOR_OPT_SEED,
            pyboolector.BTOR_OPT_VERBOSITY,
            pyboolector.BTOR_OPT_LOGLEVEL,
            pyboolector.BTOR_OPT_REWRITE_LEVEL,
            pyboolector.BTOR_OPT_SKELETON_PREPROC,
            pyboolector.BTOR_OPT_ACKERMANN,
            pyboolector.BTOR_OPT_BETA_REDUCE,
            pyboolector.BTOR_OPT_ELIMINATE_SLICES,
            pyboolector.BTOR_OPT_VAR_SUBST,
            pyboolector.BTOR_OPT_UCOPT,
            pyboolector.BTOR_OPT_MERGE_LAMBDAS,
            pyboolector.BTOR_OPT_EXTRACT_LAMBDAS,
            pyboolector.BTOR_OPT_NORMALIZE,
            pyboolector.BTOR_OPT_NORMALIZE_ADD,
            pyboolector.BTOR_OPT_FUN_PREPROP,
            pyboolector.BTOR_OPT_FUN_PRESLS,
            pyboolector.BTOR_OPT_FUN_DUAL_PROP,
            pyboolector.BTOR_OPT_FUN_DUAL_PROP_QSORT,
            pyboolector.BTOR_OPT_FUN_JUST,
            pyboolector.BTOR_OPT_FUN_JUST_HEURISTIC,
            pyboolector.BTOR_OPT_FUN_LAZY_SYNTHESIZE,
            pyboolector.BTOR_OPT_FUN_EAGER_LEMMAS,
            pyboolector.BTOR_OPT_FUN_STORE_LAMBDAS,
            pyboolector.BTOR_OPT_SLS_NFLIPS,
            pyboolector.BTOR_OPT_SLS_STRATEGY,
            pyboolector.BTOR_OPT_SLS_JUST,
            pyboolector.BTOR_OPT_SLS_MOVE_GW,
            pyboolector.BTOR_OPT_SLS_MOVE_RANGE,
            pyboolector.BTOR_OPT_SLS_MOVE_SEGMENT,
            pyboolector.BTOR_OPT_SLS_MOVE_RAND_WALK,
            pyboolector.BTOR_OPT_SLS_PROB_MOVE_RAND_WALK,
            pyboolector.BTOR_OPT_SLS_MOVE_RAND_ALL,
            pyboolector.BTOR_OPT_SLS_MOVE_RAND_RANGE,
            pyboolector.BTOR_OPT_SLS_MOVE_PROP,
            pyboolector.BTOR_OPT_SLS_MOVE_PROP_N_PROP,
            pyboolector.BTOR_OPT_SLS_MOVE_PROP_N_SLS,
            pyboolector.BTOR_OPT_SLS_MOVE_PROP_FORCE_RW,
            pyboolector.BTOR_OPT_SLS_MOVE_INC_MOVE_TEST,
            pyboolector.BTOR_OPT_SLS_USE_RESTARTS,
            pyboolector.BTOR_OPT_SLS_USE_BANDIT,
            pyboolector.BTOR_OPT_PROP_NPROPS,
            pyboolector.BTOR_OPT_PROP_USE_RESTARTS,
            pyboolector.BTOR_OPT_PROP_USE_BANDIT,
            pyboolector.BTOR_OPT_PROP_PATH_SEL,
            pyboolector.BTOR_OPT_PROP_PROB_USE_INV_VALUE,
            pyboolector.BTOR_OPT_PROP_PROB_FLIP_COND,
            pyboolector.BTOR_OPT_PROP_PROB_FLIP_COND_CONST,
            pyboolector.BTOR_OPT_PROP_FLIP_COND_CONST_DELTA,
            pyboolector.BTOR_OPT_PROP_FLIP_COND_CONST_NPATHSEL,
            pyboolector.BTOR_OPT_PROP_PROB_SLICE_KEEP_DC,
            pyboolector.BTOR_OPT_PROP_PROB_CONC_FLIP,
            pyboolector.BTOR_OPT_PROP_PROB_SLICE_FLIP,
            pyboolector.BTOR_OPT_PROP_PROB_EQ_FLIP,
            pyboolector.BTOR_OPT_PROP_PROB_AND_FLIP,
            pyboolector.BTOR_OPT_PROP_NO_MOVE_ON_CONFLICT,
            pyboolector.BTOR_OPT_AIGPROP_USE_RESTARTS,
            pyboolector.BTOR_OPT_AIGPROP_USE_BANDIT,
            pyboolector.BTOR_OPT_QUANT_SYNTH,
            pyboolector.BTOR_OPT_QUANT_DUAL_SOLVER,
            pyboolector.BTOR_OPT_QUANT_SYNTH_LIMIT,
            pyboolector.BTOR_OPT_QUANT_SYNTH_QI,
            pyboolector.BTOR_OPT_QUANT_DER,
            pyboolector.BTOR_OPT_QUANT_CER,
            pyboolector.BTOR_OPT_QUANT_MINISCOPE,
            pyboolector.BTOR_OPT_SORT_EXP,
            pyboolector.BTOR_OPT_SORT_AIG,
            pyboolector.BTOR_OPT_SORT_AIGVEC,
            pyboolector.BTOR_OPT_AUTO_CLEANUP_INTERNAL,
            pyboolector.BTOR_OPT_SIMPLIFY_CONSTRAINTS,
            pyboolector.BTOR_OPT_CHK_FAILED_ASSUMPTIONS,
            pyboolector.BTOR_OPT_CHK_MODEL,
            pyboolector.BTOR_OPT_CHK_UNCONSTRAINED,
            pyboolector.BTOR_OPT_PARSE_INTERACTIVE,
            pyboolector.BTOR_OPT_SAT_ENGINE_LGL_FORK,
            pyboolector.BTOR_OPT_SAT_ENGINE_CADICAL_FREEZE,
            pyboolector.BTOR_OPT_SAT_ENGINE_N_THREADS,
            pyboolector.BTOR_OPT_SIMP_NORMAMLIZE_ADDERS,
            pyboolector.BTOR_OPT_DECLSORT_BV_WIDTH,
            pyboolector.BTOR_OPT_QUANT_SYNTH_ITE_COMPLETE,
            pyboolector.BTOR_OPT_QUANT_FIXSYNTH,
            pyboolector.BTOR_OPT_RW_ZERO_LOWER_SLICE,
            pyboolector.BTOR_OPT_NONDESTR_SUBST,
        ]

    def _set_option(self, btor, name, value):
        available_options = {
            pyboolector.BoolectorOpt(btor, io).lng: io for io in self.internal_options
        }
        try:
            btor.Set_opt(available_options[name], value)
        except TypeError:
            raise PysmtValueError("Error setting the option '%s=%s'" % (name, value))
        except pyboolector.BoolectorException:
            raise PysmtValueError("Error setting the option '%s=%s'" % (name, value))
        except KeyError:
            raise PysmtValueError(
                "Unable to set non-existing option '%s'. "
                "The accepted options options are: %s"
                % (
                    name,
                    ", ".join(
                        pyboolector.BoolectorOpt(btor, io).lng
                        for io in self.internal_options
                    ),
                )
            )

    def __call__(self, solver):
        if self.generate_models:
            self._set_option(solver.btor, "model-gen", 1)
        else:
            self._set_option(solver.btor, "model-gen", 0)
        if self.incrementality:
            self._set_option(solver.btor, "incremental", 1)

        for k, v in self.solver_options.items():
            # Note: Options values in btor are mostly integers
            self._set_option(solver.btor, str(k), v)


# EOC BoolectorOptions


class BoolectorSolver(
    IncrementalTrackingSolver, UnsatCoreSolver, SmtLibBasicSolver, SmtLibIgnoreMixin
):

    LOGICS = [QF_BV, QF_UFBV, QF_ABV, QF_AUFBV, QF_AX]
    OptionsClass = BoolectorOptions

    def __init__(self, environment, logic, **options):
        IncrementalTrackingSolver.__init__(
            self, environment=environment, logic=logic, **options
        )
        self.btor = pyboolector.Boolector()
        self.options(self)
        self.converter = BTORConverter(environment, self.btor)
        self.mgr = environment.formula_manager
        self.declarations = {}
        self._named_assertions = {}
        return

        # EOC BoolectorOptions

        pass

    @clear_pending_pop
    def _reset_assertions(self):
        self.btor = pyboolector.Boolector()
        self.options(self)
        self.converter = BTORConverter(self.environment, self.btor)
        self.declarations = {}

    @clear_pending_pop
    def declare_variable(self, var):
        raise NotImplementedError

    @clear_pending_pop
    def _add_assertion(self, formula, named=None):
        self._assert_is_boolean(formula)
        term = self.converter.convert(formula)
        if self.options.unsat_cores_mode is None:
            self.btor.Assert(term)
        else:
            if self.options.unsat_cores_mode == "named" and named is not None:
                self._named_assertions[formula] = named
            # need to use assumptions to get unsat cores
            self.btor.Assume(term)
        return formula

    def get_model(self):
        assignment = {}
        for s, _ in self.converter.declared_vars.items():
            assignment[s] = self.get_value(s)
        return EagerModel(assignment=assignment, environment=self.environment)

    @clear_pending_pop
    def _solve(self, assumptions=None):
        if assumptions is not None:
            btor_assumptions = [self.converter.convert(a) for a in assumptions]
            self.btor.Assume(*btor_assumptions)

        res = self.btor.Sat()

        # need to re-add assumptions if in unsat-core mode
        # which uses Assume instead of Assert
        if self.options.unsat_cores_mode is not None:
            for a in self._assertion_stack:
                self._add_assertion(a)

        if res == self.btor.SAT:
            return True
        elif res == self.btor.UNSAT:
            return False
        else:
            raise SolverReturnedUnknownResultError

    def get_unsat_core(self):
        """After a call to solve() yielding UNSAT, returns the unsat core as a
        set of formulae"""
        self._check_unsat_core_config()

        if self.options.unsat_cores_mode == "all":
            unsat_core = set()
            # relies on this assertion stack being ordered
            assert isinstance(self._assertion_stack, list)
            btor_assertions = [self.converter.convert(a) for a in self._assertion_stack]
            in_unsat_core = self.btor.Failed(*btor_assertions)
            for a, in_core in zip(self._assertion_stack, in_unsat_core):
                if in_core:
                    unsat_core.add(a)
            return unsat_core
        else:
            return self.get_named_unsat_core().values()

    def get_named_unsat_core(self):
        """After a call to solve() yielding UNSAT, returns the unsat core as a
        dict of names to formulae"""
        self._check_unsat_core_config()

        if self.options.unsat_cores_mode == "named":
            unsat_core = {}
            # relies on this assertion stack being ordered
            assert isinstance(self._assertion_stack, list)
            btor_named_assertions = [
                self.converter.convert(a) for a in self._named_assertions.keys()
            ]
            in_unsat_core = self.btor.Failed(*btor_named_assertions)
            for a, in_core in zip(self._assertion_stack, in_unsat_core):
                if in_core:
                    name = self._named_assertions[a]
                    unsat_core[name] = a
            return unsat_core
        else:
            return dict(("_a%d" % i, f) for i, f in enumerate(self.get_unsat_core()))

    @clear_pending_pop
    def _push(self, levels=1):
        self.btor.Push(levels)

    @clear_pending_pop
    def _pop(self, levels=1):
        self.btor.Pop(levels)

    def print_model(self, name_filter=None):
        for var in self.declarations:
            if name_filter is None or not var.symbol_name().startswith(name_filter):
                print("%s = %s" % (var.symbol_name(), self.get_value(var)))

    def get_value(self, item):
        self._assert_no_function_type(item)
        itype = item.get_type()
        titem = self.converter.convert(item)
        if itype.is_bv_type():
            return self.mgr.BV(titem.assignment, item.bv_width())
        elif itype.is_bool_type():
            return self.mgr.Bool(bool(int(titem.assignment)))
        else:
            assert itype.is_array_type()
            assert itype.index_type.is_bv_type()
            assert itype.elem_type.is_bv_type()

            idx_width = itype.index_type.width
            val_width = itype.elem_type.width
            assign = {}
            for (idx, val) in titem.assignment:
                assign[self.mgr.BV(idx, idx_width)] = self.mgr.BV(val, val_width)
            return self.mgr.Array(itype.index_type, self.mgr.BV(0, val_width), assign)

    def _exit(self):
        del self.btor


class BTORConverter(Converter, DagWalker):
    def __init__(self, environment, btor):
        DagWalker.__init__(self, environment)
        self.mgr = environment.formula_manager
        self._get_type = environment.stc.get_type
        self._back_memoization = {}
        self._btor = btor
        self.declared_vars = {}
        self.function_declarations = {}
        return

    @catch_conversion_error
    def convert(self, formula):
        return self.walk(formula)

    def back(self, expr):
        raise NotImplementedError

    def declare_function(self, formula):
        name = formula.function_name()
        if name not in self.function_declarations:
            tp = self._type_to_btor(name.symbol_type())
            decl = self._btor.UF(tp)
            self.function_declarations[name] = decl
        else:
            decl = self.function_declarations[name]
        return decl

    def walk_and(self, formula, args, **kwargs):
        assert len(args) >= 2
        res = self._btor.And(args[0], args[1])
        for conj in args[2:]:
            res = self._btor.And(res, conj)
        return res

    def walk_or(self, formula, args, **kwargs):
        assert len(args) >= 2
        res = self._btor.Or(args[0], args[1])
        for disj in args[2:]:
            res = self._btor.Or(res, disj)
        return res

    def walk_not(self, formula, args, **kwargs):
        return self._btor.Not(args[0])

    def walk_symbol(self, formula, **kwargs):
        symbol_type = formula.symbol_type()
        if symbol_type.is_bool_type():
            res = self._btor.Var(self._btor.BitVecSort(1), formula.symbol_name())
        elif symbol_type.is_real_type():
            raise ConvertExpressionError
        elif symbol_type.is_int_type():
            raise ConvertExpressionError
        elif symbol_type.is_array_type():
            # BTOR supports only Arrays of Type (BV, BV)
            index_type = symbol_type.index_type
            elem_type = symbol_type.elem_type
            if not (index_type.is_bv_type() and elem_type.is_bv_type()):
                raise ConvertExpressionError(
                    "BTOR supports only Array(BV,BV). "
                    "Type '%s' was given." % str(symbol_type)
                )
            res = self._btor.Array(
                self._btor.ArraySort(
                    self._btor.BitVecSort(index_type.width),
                    self._btor.BitVecSort(elem_type.width),
                ),
                formula.symbol_name(),
            )
        elif symbol_type.is_bv_type():
            res = self._btor.Var(
                self._btor.BitVecSort(formula.bv_width()), formula.symbol_name()
            )
        else:
            raise SolverReturnedUnknownResultError("Unknown type for BTOR")
        self.declared_vars[formula] = res
        return res

    def walk_iff(self, formula, args, **kwargs):
        return self._btor.Iff(*args)

    def walk_implies(self, formula, args, **kwargs):
        return self._btor.Implies(*args)

    def walk_ite(self, formula, args, **kwargs):
        return self._btor.Cond(*args)

    def walk_bool_constant(self, formula, **kwargs):
        return self._btor.Const(formula.constant_value())

    def walk_equals(self, formula, args, **kwargs):
        args = self._extend_bv_equal_width(*args)
        return self._btor.Eq(*args)

    def walk_function(self, formula, args, **kwargs):
        _uf = self.declare_function(formula)
        return _uf(*args)

    def walk_bv_constant(self, formula, **kwargs):
        value = to_python_integer(formula.constant_value())
        width = formula.bv_width()
        return self._btor.Const(value, width)

    def walk_bv_ult(self, formula, args, **kwargs):
        return self._btor.Ult(args[0], args[1])

    def walk_bv_ule(self, formula, args, **kwargs):
        return self._btor.Ulte(args[0], args[1])

    def walk_bv_concat(self, formula, args, **kwargs):
        return self._btor.Concat(args[0], args[1])

    def walk_bv_extract(self, formula, args, **kwargs):
        start = formula.bv_extract_start()
        end = formula.bv_extract_end()
        return self._btor.Slice(args[0], end, start)

    def walk_bv_or(self, formula, args, **kwargs):
        return self.walk_or(formula, args, **kwargs)

    def walk_bv_not(self, formula, args, **kwargs):
        return self.walk_not(formula, args, **kwargs)

    def walk_bv_and(self, formula, args, **kwargs):
        return self.walk_and(formula, args, **kwargs)

    def walk_bv_xor(self, formula, args, **kwargs):
        return self._btor.Xor(*args)

    def walk_bv_add(self, formula, args, **kwargs):
        return self._btor.Add(args[0], args[1])

    def walk_bv_sub(self, formula, args, **kwargs):
        return self._btor.Sub(args[0], args[1])

    def walk_bv_neg(self, formula, args, **kwargs):
        return -args[0]

    def walk_bv_mul(self, formula, args, **kwargs):
        return args[0] * args[1]

    def walk_bv_udiv(self, formula, args, **kwargs):
        return args[0] / args[1]

    def walk_bv_urem(self, formula, args, **kwargs):
        return self._btor.Urem(args[0], args[1])

    def walk_bv_lshl(self, formula, args, **kwargs):
        return self._btor.Sll(args[0], args[1])

    def walk_bv_lshr(self, formula, args, **kwargs):
        return self._btor.Srl(args[0], args[1])

    def walk_bv_rol(self, formula, args, **kwargs):
        return self._btor.Rol(args[0], formula.bv_rotation_step())

    def walk_bv_ror(self, formula, args, **kwargs):
        return self._btor.Ror(args[0], formula.bv_rotation_step())

    def walk_bv_zext(self, formula, args, **kwargs):
        return self._btor.Uext(args[0], formula.bv_extend_step())

    def walk_bv_sext(self, formula, args, **kwargs):
        return self._btor.Sext(args[0], formula.bv_extend_step())

    def walk_bv_slt(self, formula, args, **kwargs):
        return self._btor.Slt(args[0], args[1])

    def walk_bv_sle(self, formula, args, **kwargs):
        return self._btor.Slte(args[0], args[1])

    def walk_bv_comp(self, formula, args, **kwargs):
        return self._btor.Eq(args[0], args[1])

    def walk_bv_sdiv(self, formula, args, **kwargs):
        return self._btor.Sdiv(args[0], args[1])

    def walk_bv_srem(self, formula, args, **kwargs):
        return self._btor.Srem(args[0], args[1])

    def walk_bv_ashr(self, formula, args, **kwargs):
        return self._btor.Sra(args[0], args[1])

    def walk_array_store(self, formula, args, **kwargs):
        return self._btor.Write(args[0], args[1], args[2])

    def walk_array_select(self, formula, args, **kwargs):
        return self._btor.Read(args[0], args[1])

    def walk_array_value(self, formula, args, **kwargs):
        raise ConvertExpressionError("btor does not support constant arrays")

    def _type_to_btor(self, tp):
        if tp.is_bool_type():
            return self._btor.BoolSort()
        elif tp.is_real_type():
            raise ConvertExpressionError
        elif tp.is_int_type():
            raise ConvertExpressionError
        elif tp.is_bv_type():
            return self._btor.BitVecSort(tp.width)
        elif tp.is_array_type():
            raise ConvertExpressionError("Unsupported Array Type")
        else:
            assert tp.is_function_type(), "Unsupported type '%s'" % tp
            stps = [self._type_to_btor(x) for x in tp.param_types]
            rtp = self._type_to_btor(tp.return_type)
            return self._btor.FunSort(stps, rtp)

    def _extend_bv_pow2(self, btor_formula, signed=False):
        """BTOR requires that many operands have width that is a power of 2"""
        w = btor_formula.width
        target_w = 2 ** int(ceil(log(w, 2)))
        # Skip if width is ok
        if target_w == w:
            return btor_formula
        if signed:
            return self._btor.Sext(btor_formula, (target_w - w))
        else:
            return self._btor.Uext(btor_formula, (target_w - w))

    def _extend_bv_equal_width(self, arg1, arg2):
        if arg1.width == arg2.width:
            return (arg1, arg2)
        elif arg1.width > arg2.width:
            ext = arg1.width - arg2.width
            return (arg1, self._btor.Uext(arg2, ext))
        elif arg1.width < arg2.width:
            ext = arg2.width - arg1.width
            return (self._btor.Uext(arg1, ext), arg2)



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

VERSION = (0, 1, 5)
if len(VERSION) == 5:
    import subprocess

    try:
        git_version = subprocess.check_output(
            ["git", "describe", "--dirty=-wip"], stderr=subprocess.STDOUT
        )
        commits_from_tag = git_version.strip().decode("ascii")
        commits_from_tag = commits_from_tag.split("-")[1]
        commits_from_tag = int(commits_from_tag)
        VERSION = VERSION[:4] + (commits_from_tag,)
    except Exception as ex:
        pass

# PEP440 Format
__version__ = "%d.%d.%d.%s%d" % VERSION if len(VERSION) == 5 else "%d.%d.%d" % VERSION



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from io import StringIO

import pysmt.logics as logics
import pysmt.smtlib.commands as smtcmd
from pysmt.shortcuts import (
    Real,
    Plus,
    Symbol,
    Equals,
    And,
    Bool,
    Or,
    Not,
    Div,
    LT,
    LE,
    Int,
    ToReal,
    Iff,
    Exists,
    Times,
    FALSE,
    BVLShr,
    BVLShl,
    BVAShr,
    BV,
    BVAdd,
    BVULT,
    BVMul,
    Select,
    Array,
    Ite,
    String,
    Function,
    to_smtlib,
    ForAll,
)
from pysmt.shortcuts import Solver, get_env, qelim, get_model, TRUE, ExactlyOne
from pysmt.typing import REAL, BOOL, INT, BVType, FunctionType, ArrayType
from pysmt.test import (
    TestCase,
    skipIfSolverNotAvailable,
    skipIfNoSolverForLogic,
    skipIfNoQEForLogic,
)
from pysmt.test import main
from pysmt.exceptions import (
    ConvertExpressionError,
    PysmtValueError,
    PysmtTypeError,
    InternalSolverError,
)
from pysmt.test.examples import get_example_formulae
from pysmt.environment import Environment
from pysmt.rewritings import cnf_as_set
from pysmt.smtlib.parser import SmtLibParser
from pysmt.smtlib.commands import DECLARE_FUN
from pysmt.smtlib.script import SmtLibCommand
from pysmt.logics import get_closer_smtlib_logic
from pysmt.constants import Fraction


class TestRegressions(TestCase):
    @skipIfSolverNotAvailable("msat")
    @skipIfSolverNotAvailable("z3")
    def test_plus_converts_correctly_n_ary_functions(self):
        """Handling of Plus n-ary functionality.

        Only the first two elements were translated to the solver
        """

        a = Symbol("a", REAL)
        b = Symbol("b", REAL)
        c = Symbol("c", REAL)

        p1 = Plus(
            a,
            Real((1, 6)),
            b,
            c,
        )
        p2 = Plus(a, b, c, Real((1, 6)))

        self.assertValid(Equals(p1, p2))
        self.assertValid(Equals(p1, p2), solver_name="z3")
        self.assertValid(Equals(p1, p2), solver_name="msat")

    def test_substitute_memoization(self):
        a = Symbol("A", BOOL)
        b = Symbol("B", BOOL)

        f = And(a, b)
        g = f.substitute({a: Bool(True)})
        h = f.substitute({a: Bool(False)})

        self.assertNotEqual(h, g)

    @skipIfSolverNotAvailable("msat")
    def test_msat_bool_back_conversion(self):
        f = Symbol("A")
        with Solver(name="msat") as solver:
            solver.solve()
            val = solver.get_value(Symbol("A"))
            self.assertTrue(val.is_bool_constant())

    @skipIfSolverNotAvailable("msat")
    @skipIfSolverNotAvailable("z3")
    def test_conversion_of_fractions_in_z3(self):
        self.assertValid(
            Equals(Real(Fraction(1, 9)), Div(Real(1), Real(9))), solver_name="msat"
        )
        self.assertValid(
            Equals(Real(Fraction(1, 9)), Div(Real(1), Real(9))), solver_name="z3"
        )

    def test_simplifying_int_plus_changes_type_of_expression(self):
        varA = Symbol("At", INT)
        varB = Symbol("Bt", INT)
        get_type = get_env().stc.get_type
        f = Plus(varB, Int(1))
        old_type = get_type(f)
        f = f.simplify()
        new_type = get_type(f)
        self.assertEqual(new_type, old_type)

    @skipIfNoSolverForLogic(logics.QF_UFLIRA)
    def test_nary_operators_in_solver_converter(self):
        """Conversion of n-ary operators was not handled correctly by converters."""
        x = Symbol("x")
        r = Symbol("p", REAL)
        f_and_one = And(x)
        f_or_one = Or(x)
        f_plus_one = LT(Plus(r), Real(0))

        ten_x = [x, x, x, x, x, x, x, x, x, x]
        f_and_many = And(ten_x)
        f_or_many = Or(ten_x)
        f_plus_many = LT(Plus(r, r, r, r, r, r, r, r, r, r, r), Real(0))

        for name in get_env().factory.all_solvers(logic=logics.QF_BOOL):
            self.assertSat(f_and_one, solver_name=name)
            self.assertSat(f_or_one, solver_name=name)
            self.assertSat(f_and_many, solver_name=name)
            self.assertSat(f_or_many, solver_name=name)

        for name in get_env().factory.all_solvers(logic=logics.QF_UFLIRA):
            self.assertSat(f_plus_one, solver_name=name)
            self.assertSat(f_plus_many, solver_name=name)

    def test_dependencies_not_includes_toreal(self):
        p = Symbol("p", INT)
        r = ToReal(p)
        deps = r.get_free_variables()

        self.assertIn(p, deps)
        self.assertNotIn(r, deps)

    def test_infix_notation_wrong_le(self):
        p = Symbol("p", INT)
        get_env().enable_infix_notation = True
        self.assertEqual(LE(p, Int(2)), p <= Int(2))

    def test_multiple_declaration_w_same_functiontype(self):
        ft1 = FunctionType(REAL, [REAL])
        ft2 = FunctionType(REAL, [REAL])

        f1 = Symbol("f1", ft1)
        # The following raises an exception if not (ft1 == ft2)
        # since the same symbol has already been defined with
        # a "different" type.
        f1 = Symbol("f1", ft2)

    @skipIfSolverNotAvailable("z3")
    def test_z3_iff(self):
        z3 = Solver(name="z3")
        conv = z3.converter

        x, y = Symbol("x"), Symbol("y")
        term = conv.convert(Iff(x, y))
        back = conv.back(term)
        self.assertEqual(Iff(x, y), back)

    @skipIfSolverNotAvailable("msat")
    def test_msat_iff(self):
        msat = Solver(name="msat")
        conv = msat.converter

        x, y = Symbol("x"), Symbol("y")
        term = conv.convert(Iff(x, y))
        back = conv.back(term)
        # Mathsat can reorder variables...
        self.assertTrue(Iff(x, y) == back or Iff(y, x) == back)

    def test_multiple_exit(self):
        for sname in get_env().factory.all_solvers():
            # Multiple exits should be ignored
            s = Solver(name=sname)
            s.exit()
            s.exit()
            self.assertTrue(True)

    @skipIfNoQEForLogic(logics.LIA)
    def test_lia_qe_requiring_modulus(self):
        x = Symbol("x", INT)
        y = Symbol("y", INT)
        f = Exists([x], Equals(y, Times(x, Int(2))))
        with self.assertRaises(ConvertExpressionError):
            qelim(f)

        try:
            qelim(f)
        except ConvertExpressionError as ex:
            # The modulus operator must be there
            self.assertTrue(
                "%2" in str(ex.expression) or "int_mod_congr" in str(ex.expression)
            )

    @skipIfSolverNotAvailable("msat")
    def test_msat_partial_model(self):
        msat = Solver(name="msat")
        x, y = Symbol("x"), Symbol("y")
        msat.add_assertion(x)
        c = msat.solve()
        self.assertTrue(c)

        model = msat.get_model()
        self.assertNotIn(y, model)
        self.assertIn(x, model)
        msat.exit()

    @skipIfSolverNotAvailable("z3")
    def test_z3_model_iteration(self):
        x, y = Symbol("x"), Symbol("y")
        m = get_model(And(x, y), solver_name="z3")
        self.assertIsNotNone(m)

        for _, v in m:
            self.assertEqual(v, TRUE())

    def test_exactlyone_w_generator(self):
        x, y = Symbol("x"), Symbol("y")

        elems = [x, y]
        f1 = ExactlyOne(elems)
        f2 = ExactlyOne(e for e in elems)

        self.assertEqual(f1, f2)

    def test_determinism(self):
        def get_set(env):
            mgr = env.formula_manager
            r = set(mgr.Symbol("x%d" % i) for i in range(1000))
            for (f, _, _, _) in get_example_formulae(env):
                r |= set([f])
            return r

        # As first thing on the environment we build the set of formulae
        l1 = list(get_set(get_env()))

        # We try this ten times...
        for _ in range(10):
            # Do something to screw up memory layout...
            for y in (Symbol("y%d" % i) for i in range(1000)):
                self.assertIsNotNone(y)

            with Environment() as new_env:
                # As first thing on the environment we build the set of formulae
                l_test = list(get_set(new_env))

                # The ordering of the sets should be the same...
                for i, f in enumerate(l1):
                    nf = new_env.formula_manager.normalize(f)
                    self.assertEqual(nf, l_test[i])

    def test_is_one(self):
        self.assertTrue(Int(1).is_one())
        self.assertTrue(Real(1).is_one())
        self.assertTrue(Int(0).is_zero())
        self.assertTrue(Real(0).is_zero())

    def test_cnf_as_set(self):
        r = cnf_as_set(Symbol("x"))
        self.assertTrue(type(r) == frozenset)

    def test_substitute_to_real(self):
        p = Symbol("p", INT)
        f = LT(ToReal(p), Real(0))

        new_f = f.substitute({p: Real(1)}).simplify()
        self.assertEqual(new_f, Bool(False))

    def test_empty_string_symbol(self):
        with self.assertRaises(PysmtValueError):
            Symbol("")

    def test_smtlib_info_quoting(self):
        cmd = SmtLibCommand(smtcmd.SET_INFO, [":source", "This\nis\nmultiline!"])
        output = cmd.serialize_to_string()
        self.assertEqual(output, "(set-info :source |This\nis\nmultiline!|)")

    def test_parse_define_fun(self):
        smtlib_input = (
            "(declare-fun z () Bool)" "(define-fun .def_1 ((z Bool)) Bool (and z z))"
        )
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        parser.get_script(buffer_)

    def test_parse_define_fun_bind(self):
        smtlib_input = (
            "(declare-fun y () Bool)" "(define-fun .def_1 ((z Bool)) Bool (and z z))"
        )
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        parser.get_script(buffer_)

    def test_parse_bvx_var(self):
        """bvX is a valid identifier."""
        smtlib_input = """
        (declare-fun bv1 () (_ BitVec 8))
        (assert (bvult (_ bv0 8) (bvmul (bvadd bv1 (_ bv1 8)) (_ bv5 8))))
        (check-sat)"""
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        script = parser.get_script(buffer_)
        # Check Parsed result
        iscript = iter(script)
        cmd = next(iscript)
        self.assertEqual(cmd.name, DECLARE_FUN)
        bv1 = cmd.args[0]
        self.assertEqual(bv1.symbol_type().width, 8)
        cmd = next(iscript)
        parsed_f = cmd.args[0]
        target_f = BVULT(BV(0, 8), BVMul(BVAdd(bv1, BV(1, 8)), BV(5, 8)))
        self.assertEqual(parsed_f, target_f)

    def test_simplify_times(self):
        a, b = Real(5), Real((1, 5))
        f = Times(a, b).simplify()
        self.assertEqual(f.constant_value(), 1)

    @skipIfSolverNotAvailable("yices")
    def test_yices_push(self):
        with Solver(name="yices") as solver:
            solver.add_assertion(FALSE())
            res = solver.solve()
            self.assertFalse(res)
            solver.push()
            solver.add_assertion(TRUE())
            res = solver.solve()
            self.assertFalse(res)
            solver.pop()

    def test_qf_bool_smt2(self):
        # QF_BOOL does not exist in SMT-LIB
        # This test is to enforce the consistent choice of QF_UF
        close_l = get_closer_smtlib_logic(logics.QF_BOOL)
        self.assertEqual(close_l, logics.QF_UF)
        # For BOOL we use LRA
        close_l = get_closer_smtlib_logic(logics.BOOL)
        self.assertEqual(close_l, logics.LRA)

    def test_exactly_one_unpacking(self):
        s1, s2 = Symbol("x"), Symbol("y")
        f1 = ExactlyOne((s for s in [s1, s2]))
        f2 = ExactlyOne([s1, s2])
        f3 = ExactlyOne(s1, s2)

        self.assertEqual(f1, f2)
        self.assertEqual(f2, f3)

    @skipIfSolverNotAvailable("btor")
    def test_btor_bitwidth_bug_in_shift(self):
        # (384, 384, 9)
        # (x69 >> 1_384)
        s = Solver(name="btor")
        x69 = Symbol("x69", BVType(384))
        # BVLShr
        f = BVLShr(x69, BV(1, 384))
        c = s.converter.convert(f)
        self.assertIsNotNone(c)
        # BVLShl
        f = BVLShl(x69, BV(1, 384))
        c = s.converter.convert(f)
        self.assertIsNotNone(c)
        # BVAShr
        f = BVAShr(x69, BV(1, 384))
        c = s.converter.convert(f)
        self.assertIsNotNone(c)

    @skipIfSolverNotAvailable("btor")
    def test_btor_get_non_bool_value(self):
        with Solver(name="btor") as s:
            x = Symbol("x", BVType(16))
            s.add_assertion(Equals(x, BV(1, 16)))
            self.assertTrue(s.solve())
            self.assertEqual(s.get_value(Equals(x, BV(1, 16))), TRUE())
            self.assertEqual(s.get_value(BVAdd(x, BV(1, 16))), BV(2, 16))

    @skipIfSolverNotAvailable("btor")
    def test_btor_get_array_element(self):
        with Solver(name="btor") as s:
            x = Symbol("a", ArrayType(BVType(16), BVType(16)))
            s.add_assertion(Equals(Select(x, BV(1, 16)), BV(1, 16)))
            s.add_assertion(Equals(Select(x, BV(2, 16)), BV(3, 16)))
            self.assertTrue(s.solve())
            self.assertEqual(s.get_value(Select(x, BV(1, 16))), BV(1, 16))
            self.assertIsNotNone(s.get_value(x))

    def test_smtlib_define_fun_serialization(self):
        smtlib_input = "(define-fun init ((x Bool)) Bool (and x (and x (and x (and x (and x (and x x)))))))"
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        s = parser.get_script(buffer_)
        for c in s:
            res = c.serialize_to_string(daggify=False)
        self.assertEqual(res.replace("__x0", "x"), smtlib_input)

    @skipIfSolverNotAvailable("z3")
    def test_z3_nary_back(self):
        from z3 import Tactic

        r = Symbol("r", REAL)
        s = Symbol("s", REAL)
        t = Symbol("t", REAL)
        f = Equals(Times(r, s, t), Real(0))

        with Solver(name="z3") as solver:
            z3_f = solver.converter.convert(f)
            z3_f = Tactic("simplify", solver.z3.ctx)(z3_f).as_expr()
            fp = solver.converter.back(z3_f)
            self.assertValid(Iff(f, fp), (f, fp))

    def test_array_initialization_printing(self):
        self.assertEqual(
            str(Array(INT, Int(0), {Int(1): Int(2)})), "Array{Int, Int}(0)[1 := 2]"
        )

    @skipIfSolverNotAvailable("btor")
    def test_boolector_assumptions(self):
        with Solver(name="btor") as solver:
            x = Symbol("x")
            y = Symbol("y")
            solver.add_assertion(Or(x, y))
            solver.solve([Not(x), Not(y)])
            btor_notx = solver.converter.convert(Not(x))
            btor_noty = solver.converter.convert(Not(y))
            self.assertEqual(solver.btor.Failed(btor_notx, btor_noty), [True, True])

    def test_parse_declare_const(self):
        smtlib_input = """
        (declare-const s Int)
        (check-sat)"""
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        script = parser.get_script(buffer_)
        self.assertIsNotNone(script)

    @skipIfSolverNotAvailable("z3")
    def test_z3_conversion_ite(self):
        with Solver(name="z3") as solver:
            x = Symbol("x")
            y = Symbol("y")
            f = Ite(x, y, FALSE())
            solver.add_assertion(f)
            self.assertTrue(solver.solve())

    def test_parse_exception(self):
        from pysmt.exceptions import PysmtSyntaxError

        smtlib_input = "(declare-const x x x Int)" + "(check-sat)"
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        try:
            parser.get_script(buffer_)
            self.assertFalse(True)
        except PysmtSyntaxError as ex:
            self.assertEqual(ex.pos_info[0], 0)
            self.assertEqual(ex.pos_info[1], 19)

    def test_parse_bvconst_width(self):
        smtlib_input = "(assert (> #x10 #x10))"
        parser = SmtLibParser()
        buffer_ = StringIO(smtlib_input)
        expr = parser.get_script(buffer_).get_last_formula()
        const = expr.args()[0]
        self.assertEqual(const.bv_width(), 8, const.bv_width())

    def test_equality_typing(self):
        x = Symbol("x", BOOL)
        y = Symbol("y", BOOL)
        with self.assertRaises(PysmtTypeError):
            Equals(x, y)
        with self.assertRaises(PysmtTypeError):
            LE(x, y)

    def test_string_constant_quote_escaping_hr_printer(self):
        x = String('a"b')
        y = String('""')
        z = String('"')
        self.assertEqual('"a""b"', str(x))
        self.assertEqual('""""""', str(y))
        self.assertEqual('""""', str(z))

    def test_string_constant_quote_escaping_smtlib_printer(self):
        x = String('a"b')
        y = String('""')
        z = String('"')
        self.assertEqual('"a""b"', x.to_smtlib())
        self.assertEqual('""""""', y.to_smtlib())
        self.assertEqual('""""', z.to_smtlib())

    def test_string_constant_quote_escaping_parsing(self):
        script = """
        (declare-const x String)
        (assert (= x "a""b"))
        (assert (= x "\"\""))
        (assert (= x "\"\"\"\""))
        """

        p = SmtLibParser()
        buffer = StringIO(script)
        s = p.get_script(buffer)
        self.assertEqual('a"b', s.commands[1].args[0].arg(1).constant_value())
        self.assertEqual('"', s.commands[2].args[0].arg(1).constant_value())
        self.assertEqual('""', s.commands[3].args[0].arg(1).constant_value())

    def test_pysmt_syntax_error(self):
        from pysmt.exceptions import PysmtSyntaxError

        try:
            raise PysmtSyntaxError("'define-fun' expected", (5, 5))
        except PysmtSyntaxError as ex:
            self.assertEqual(str(ex), "Line 5, Col 5: 'define-fun' expected")

    def test_function_smtlib_print(self):
        f_t = FunctionType(BOOL, [BOOL])
        f0 = Symbol("f 0", f_t)
        f0_of_false = Function(f0, [Bool(False)])
        s = to_smtlib(f0_of_false, False)
        self.assertEqual(s, "(|f 0| false)")

    @skipIfSolverNotAvailable("yices")
    def test_yices_bv_overflow(self):
        smt_script = """
        (set-logic QF_BV)
        (declare-fun s0 () (_ BitVec 64))
        (define-fun s1 () (_ BitVec 64) #xFFFFFFFFFFFFFFFF)
        (define-fun s2 () Bool (= s1 s0))
        (assert s2)
        (check-sat)
        """
        from pysmt.smtlib.parser import get_formula_strict

        f = get_formula_strict(StringIO(smt_script))
        self.assertSat(f, solver_name="yices")

    @skipIfSolverNotAvailable("yices")
    def test_yices_bv_no_overflow(self):
        smt_script = """
        (set-logic QF_BV)
        (declare-fun s0 () (_ BitVec 64))
        (define-fun s1 () (_ BitVec 64) #x7FFFFFFFFFFFFFFF)
        (define-fun s2 () Bool (= s1 s0))
        (assert s2)
        (check-sat)
        """
        from pysmt.smtlib.parser import get_formula_strict

        f = get_formula_strict(StringIO(smt_script))
        self.assertSat(f, solver_name="yices")

    def test_get_atoms_array_select(self):
        a = Symbol("a", ArrayType(INT, BOOL))
        x = Symbol("x", INT)
        p = Symbol("p", BOOL)

        phi = And(Iff(Select(a, x), p), Equals(x, Int(1)))

        atoms = phi.get_atoms()

        self.assertEqual(len(atoms), 3)
        self.assertIn(Select(a, x), atoms)
        self.assertIn(p, atoms)
        self.assertIn(Equals(x, Int(1)), atoms)

    @skipIfSolverNotAvailable("yices")
    def test_yices_quantifier(self):
        x = Symbol("x", REAL)
        f = ForAll([x], LE(x, Real(0)))
        with self.assertRaises(InternalSolverError):
            with Solver(name="yices") as s:
                s.add_assertion(f)
                self.assertFalse(s.solve())


if __name__ == "__main__":
    main()



#
# This file is part of pySMT.
#
#   Copyright 2015 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from pysmt.test import TestCase, skipIfNoSolverForLogic
from pysmt.shortcuts import (
    Not,
    Implies,
    Equals,
    Symbol,
    GE,
    GT,
    LT,
    And,
    Int,
    Plus,
    TRUE,
    FALSE,
)
from pysmt.shortcuts import (
    String,
    StrConcat,
    StrLength,
    StrContains,
    StrIndexOf,
    StrReplace,
    StrSubstr,
    StrPrefixOf,
    StrSuffixOf,
    StrToInt,
    IntToStr,
    StrCharAt,
    Solver,
)
from pysmt.typing import INT, STRING
from pysmt.logics import QF_SLIA


class TestString(TestCase):
    # MG: This test suit overlaps with examples.py
    #    we might want to include tests of more things like:
    #    - Simplifications at construction time
    #    - Infix notation
    #    - Constants and unicode support

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_length(self):
        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)
        f = Not(Implies(Equals(s1, s2), Equals(StrLength(s2), StrLength(s1))))
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_concat(self):
        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)

        f = Equals(StrConcat(s1, s1), s2)
        self.assertSat(f)
        f = Not(
            And(
                GE(StrLength(StrConcat(s1, s2)), StrLength(s1)),
                GE(StrLength(StrConcat(s1, s2)), StrLength(s2)),
            )
        )
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_contains(self):
        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)
        f = Not(Implies(And(StrContains(s1, s2), StrContains(s2, s1)), Equals(s1, s2)))
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_indexof(self):
        s1 = String("Hello World")
        t1 = String("o")
        # MG: Make offset argument optional (default 0) StrIndexOf
        f = Not(Equals(StrIndexOf(s1, t1, Int(0)), Int(4)))
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_replace(self):
        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)
        s3 = Symbol("s3", STRING)

        f = Equals(StrReplace(s1, s2, s3), s3)
        self.assertSat(f)
        f = And(
            GT(StrLength(s1), Int(0)),
            GT(StrLength(s2), Int(0)),
            GT(StrLength(s3), Int(0)),
            Not(StrContains(s1, s2)),
            Not(StrContains(s1, s3)),
            Not(Equals(StrReplace(StrReplace(s1, s2, s3), s3, s2), s1)),
        )
        self.assertUnsat(f)

        # Replace first v Replace First
        f = Implies(
            And(
                Equals(s1, String("Hello")),
                Equals(s2, StrReplace(s1, String("l"), String(" "))),
            ),
            Equals(s2, String("He lo")),
        )
        self.assertValid(f, logic="QF_SLIA")

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_substr(self):
        s1 = Symbol("s1", STRING)
        i = Symbol("index", INT)
        f = And(
            GT(i, Int(0)),
            GT(StrLength(s1), Int(1)),
            LT(i, StrLength(s1)),
            Equals(
                StrConcat(
                    StrSubstr(s1, Int(0), i),
                    StrSubstr(s1, Plus(i, Int(1)), StrLength(s1)),
                ),
                s1,
            ),
        )
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_prefixof(self):
        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)
        f = And(
            GT(StrLength(s1), Int(2)),
            GT(StrLength(s2), StrLength(s1)),
            And(StrPrefixOf(s2, s1), StrContains(s2, s1)),
        )
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_suffixof(self):
        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)
        f = And(
            GT(StrLength(s1), Int(2)),
            GT(StrLength(s2), StrLength(s1)),
            And(StrSuffixOf(s2, s1), StrContains(s2, s1)),
        )
        self.assertUnsat(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_to_int(self):
        f = Equals(StrToInt(String("1")), Int(1))
        self.assertValid(f)
        f = Equals(StrToInt(String("-55")), Int(-1))
        self.assertValid(f)
        f = Equals(StrToInt(String("pippo")), Int(-1))
        self.assertValid(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_int_to_str(self):
        f = Equals((IntToStr(Int(1))), String("1"))
        self.assertValid(f)
        f = Equals(IntToStr(Int(-1)), String(""))
        self.assertValid(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_str_charat(self):
        s1 = String("Hello")
        f = Equals(StrCharAt(s1, Int(0)), String("H"))
        self.assertValid(f)

    @skipIfNoSolverForLogic(QF_SLIA)
    def test_model(self):
        s = Symbol("s", STRING)
        f = [
            Equals(StrLength(s), Int(5)),
            Equals(StrCharAt(s, Int(0)), String("A")),
            Not(Equals(StrCharAt(s, Int(2)), String("B"))),
        ]
        with Solver(logic=QF_SLIA) as solver:
            solver.add_assertion(And(f))
            res = solver.solve()
            self.assertTrue(res)
            s_value = solver.get_value(s)
            py_value = s_value.constant_value()
            self.assertEqual(py_value[0], "A")
            self.assertNotEqual(py_value[1], "B")
            self.assertEqual(len(py_value), 5)

    def test_simplification(self):
        constA, constB = String("Hello"), String("World")
        test_set = [
            (StrLength(constA), Int(5)),
            (StrConcat(constA, constB), String("HelloWorld")),
            (StrContains(constA, String("H")), TRUE()),
            (StrContains(constB, String("H")), FALSE()),
            (StrIndexOf(constA, String("e"), Int(0)), Int(1)),
            (StrReplace(constA, String("l"), String(" ")), String("He lo")),
            (StrSubstr(constA, Int(1), Int(2)), String("el")),
            (StrPrefixOf(constA, constB), FALSE()),
            (StrPrefixOf(String("He"), constA), TRUE()),
            (StrSuffixOf(constA, constB), FALSE()),
            (StrSuffixOf(String("lo"), constB), FALSE()),
            (StrToInt(constA), Int(-1)),
            (StrToInt(String("55")), Int(55)),
            (IntToStr(Int(10)), String("10")),
            (IntToStr(Int(-1)), String("")),
            (StrCharAt(constA, Int(2)), String("l")),
        ]

        for (f, simplified) in test_set:
            self.assertEqual(f.simplify(), simplified)

    def test_theory_oracle(self):
        from pysmt.oracles import get_logic

        s1 = Symbol("s1", STRING)
        s2 = Symbol("s2", STRING)

        f = Equals(StrConcat(s1, s1), s2)
        theory = get_logic(f).theory
        self.assertTrue(theory.strings, theory)

        f = Not(
            And(
                GE(StrLength(StrConcat(s1, s2)), StrLength(s1)),
                GE(StrLength(StrConcat(s1, s2)), StrLength(s2)),
            )
        )
        theory = get_logic(f).theory
        self.assertTrue(theory.strings, theory)

        f = And(
            GT(StrLength(s1), Int(2)),
            GT(StrLength(s2), StrLength(s1)),
            And(StrSuffixOf(s2, s1), StrContains(s2, s1)),
        )
        theory = get_logic(f).theory
        self.assertTrue(theory.strings, theory)


if __name__ == "__main__":
    from pysmt.test import main

    main()



# Copyright 2014 Andrea Micheli and Marco Gario
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import absolute_import

import os
import glob

from pysmt.cmd.installers.base import SolverInstaller


class Z3Installer(SolverInstaller):

    SOLVER = "z3"

    def __init__(
        self,
        install_dir,
        bindings_dir,
        solver_version,
        mirror_link=None,
        osx=None,
        git_version=None,
        commit=None,
    ):
        arch = self.architecture
        if arch == "x86_64":
            arch = "x64"

        system = self.os_name
        if system == "linux":
            system = "ubuntu-16.04"
        elif system == "darwin":
            system = "osx-%s" % osx
        elif system == "windows":
            system = "win"

        if git_version is None:
            # Stable versions template
            if commit:
                archive_name = "z3-%s.%s-%s-%s.zip" % (
                    solver_version,
                    commit,
                    arch,
                    system,
                )
            else:
                archive_name = "z3-%s-%s-%s.zip" % (solver_version, arch, system)
            native_link = (
                "https://github.com/Z3Prover/z3/releases/download/z3-"
                + solver_version
                + "/{archive_name}"
            )
            # print(native_link)
        else:
            # Nightly build template
            archive_name = "z3-%s.%s-%s-%s.zip" % (
                solver_version,
                git_version,
                arch,
                system,
            )
            native_link = "https://github.com/pysmt/Z3bin/blob/master/nightly/{archive_name}?raw=true"

        SolverInstaller.__init__(
            self,
            install_dir=install_dir,
            bindings_dir=bindings_dir,
            solver_version=solver_version,
            archive_name=archive_name,
            native_link=native_link,
            mirror_link=mirror_link,
        )

    def move(self):
        bpath = os.path.join(self.extract_path, "bin")
        libfiles = []
        if self.os_name == "linux":
            libfiles += glob.glob(bpath + "/*.so")
        elif self.os_name == "darwin":
            libfiles += glob.glob(bpath + "/*.a")
            libfiles += glob.glob(bpath + "/*.dylib")
        elif self.os_name == "windows":
            libfiles += glob.glob(bpath + "/*.dll")
            libfiles += glob.glob(bpath + "/*.lib")

        SolverInstaller.mv(os.path.join(bpath, "python/z3"), self.bindings_dir)

        # z3 will check for shared libraries in z3/lib, before builtins.Z3_LIB_DIRS,
        # and Z3_LIBRARY_PATH env var lookup (or OS-level LD_LIBRARY_PATH)
        # (see z3/z3core.py)
        libpath = os.path.join(self.bindings_dir, "z3/lib/")
        if not os.path.exists(libpath):
            os.mkdir(libpath)

        for f in libfiles:
            SolverInstaller.mv(f, libpath)

    def get_installed_version(self):
        return self.get_installed_version_script(self.bindings_dir, "z3")



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

# Modified by Gaëtan MARGUERITTE to adapt the program to ProgSynth SMT format.
"""Defines constants for the commands of the SMT-LIB"""

ASSERT = "assert"
CHECK_SAT = "check-sat"
CHECK_SAT_ASSUMING = "check-sat-assuming"
DECLARE_CONST = "declare-const"
DECLARE_FUN = "declare-fun"
DECLARE_SORT = "declare-sort"
DEFINE_FUN = "define-fun"
DEFINE_FUN_REC = "define-fun-rec"
DEFINE_FUNS_REC = "define-funs-rec"
DEFINE_SORT = "define-sort"
ECHO = "echo"
EXIT = "exit"
GET_ASSERTIONS = "get-assertions"
GET_ASSIGNMENT = "get-assignment"
GET_INFO = "get-info"
GET_MODEL = "get-model"
GET_OPTION = "get-option"
GET_PROOF = "get-proof"
GET_UNSAT_ASSUMPTIONS = "get-unsat-assumptions"
GET_UNSAT_CORE = "get-unsat-core"
GET_VALUE = "get-value"
POP = "pop"
PUSH = "push"
RESET = "reset"
RESET_ASSERTIONS = "reset-assertions"
SET_INFO = "set-info"
SET_LOGIC = "set-logic"
SET_OPTION = "set-option"
SYNTH_FUN = "synth-fun"
CONSTRAINT_PBE = "constraint-pbe"
CONSTRAINT_SYNTAX = "constraint-syntax"
CONSTRAINT = "constraint"
CHECK_PROGSYNTH = "check-progsynth"
CHECK_SYNTH = "check-synth"
SOLUTION_PBE = "solution-pbe"
DEFINE_CONST = "define-const"
DECLARE_VAR = "declare-var"
#

SMT_LIB_2_0 = [
    SET_LOGIC,
    SET_OPTION,
    SET_INFO,
    DECLARE_SORT,
    DEFINE_SORT,
    DECLARE_FUN,
    DEFINE_FUN,
    PUSH,
    POP,
    ASSERT,
    CHECK_SAT,
    GET_ASSERTIONS,
    GET_VALUE,
    GET_MODEL,
    GET_PROOF,
    GET_UNSAT_CORE,
    GET_INFO,
    GET_OPTION,
    EXIT,
]

SMT_LIB_2_5 = SMT_LIB_2_0 + [
    CHECK_SAT_ASSUMING,
    DECLARE_CONST,
    DEFINE_FUN_REC,
    DEFINE_FUNS_REC,
    ECHO,
    GET_ASSIGNMENT,
    GET_UNSAT_ASSUMPTIONS,
    RESET,
    RESET_ASSERTIONS,
]

PROGSYNTH = [
    SYNTH_FUN,
    CONSTRAINT_PBE,
    CONSTRAINT_SYNTAX,
    CHECK_PROGSYNTH,
    SOLUTION_PBE,
    DEFINE_CONST,
    DECLARE_VAR,
    CONSTRAINT,
    CHECK_SYNTH,
    DEFINE_FUN,
]
ALL_COMMANDS = SMT_LIB_2_5 + PROGSYNTH



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import os

from io import StringIO

from pysmt.shortcuts import reset_env
from pysmt.test import TestCase
from pysmt.smtlib.parser import SmtLibParser


class TestSmtLibParserGriggio(TestCase):
    def test_griggio(self):
        for file_id in range(1, 7):
            # Parse
            script = self.parse(file_id)
            # Check cmds
            for i, cmd in enumerate(script):
                self.assertEqual(
                    cmd.name,
                    TESTS[file_id][i],
                    "Test %d: %s != %s " % (file_id, cmd.name, TESTS[file_id][i]),
                )
            # Serialize
            buf = StringIO()
            script.serialize(buf)
            self.assertTrue(True)

    def parse(self, file_id):
        fname = SMTLIB_FILE_PATTERN % file_id
        reset_env()
        parser = SmtLibParser()
        script = parser.get_script_fname(fname)
        self.assertIsNotNone(script)
        return script


SMTLIB_FILE_PATTERN = "pysmt/test/smtlib/griggio/test%d.smt2.bz2"

TESTS = {
    1: [
        "declare-fun",
        "set-option",
        "declare-sort",
        "declare-fun",
        "declare-fun",
        "assert",
        "assert",
        "push",
        "define-fun",
        "assert",
        "check-sat",
        "pop",
        "define-fun",
        "exit",
    ],
    2: [
        "push",
        "declare-sort",
        "pop",
        "declare-fun",
        "exit",
    ],
    3: [
        "set-option",
        "declare-fun",
        "declare-fun",
        "assert",
        "assert",
        "assert",
        "check-sat",
        "get-unsat-core",
        "exit",
    ],
    4: [
        "set-option",
        "declare-fun",
        "declare-fun",
        "assert",
        "assert",
        "assert",
        "check-sat",
        "get-assignment",
        "exit",
    ],
    5: [
        "set-option",
        "declare-fun",
        "declare-fun",
        "declare-fun",
        "assert",
        "assert",
        "assert",
        "assert",
        "check-sat",
        "get-value",
        "exit",
    ],
    6: [
        "declare-sort",
        "define-sort",
        "declare-fun",
        "declare-fun",
        "assert",
        "check-sat",
        "exit",
    ],
}



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import absolute_import

from pysmt.exceptions import SolverAPINotFound

try:
    import z3
except ImportError:
    raise SolverAPINotFound

# Keep array models expressed as values instead of Lambdas
# (see https://github.com/Z3Prover/z3/issues/1769)
z3.set_param("model.compact", False)


import pysmt.typing as types
import pysmt.operators as op
from pysmt.solvers.solver import (
    IncrementalTrackingSolver,
    UnsatCoreSolver,
    Model,
    Converter,
    SolverOptions,
)
from pysmt.solvers.smtlib import SmtLibBasicSolver, SmtLibIgnoreMixin
from pysmt.solvers.qelim import QuantifierEliminator

from pysmt.walkers import DagWalker
from pysmt.exceptions import (
    SolverReturnedUnknownResultError,
    SolverNotConfiguredForUnsatCoresError,
    SolverStatusError,
    ConvertExpressionError,
    UndefinedSymbolError,
    PysmtValueError,
)
from pysmt.decorators import clear_pending_pop, catch_conversion_error
from pysmt.logics import LRA, LIA, QF_UFLRA, PYSMT_LOGICS
from pysmt.oracles import get_logic
from pysmt.constants import Fraction, Numeral, is_pysmt_integer, to_python_integer


# patch z3api
z3.is_ite = lambda x: z3.is_app_of(x, z3.Z3_OP_ITE)
z3.is_function = lambda x: z3.is_app_of(x, z3.Z3_OP_UNINTERPRETED)
z3.is_array_store = lambda x: z3.is_app_of(x, z3.Z3_OP_STORE)
z3.get_payload = lambda node, i: z3.Z3_get_decl_int_parameter(
    node.ctx.ref(), node.decl().ast, i
)


class AstRefKey:
    def __init__(self, n):
        self.n = n

    def __hash__(self):
        return self.n.hash()

    def __eq__(self, other):
        return self.n.eq(other.n)


def askey(n):
    assert isinstance(n, z3.AstRef)
    return AstRefKey(n)


class Z3Model(Model):
    def __init__(self, environment, z3_model):
        Model.__init__(self, environment)
        self.z3_model = z3_model
        self.converter = Z3Converter(environment, z3_model.ctx)

    def get_value(self, formula, model_completion=True):
        titem = self.converter.convert(formula)
        z3_res = self.z3_model.eval(titem, model_completion=model_completion)
        return self.converter.back(z3_res, model=self.z3_model)

    def iterator_over(self, language):
        for x in language:
            yield x, self.get_value(x, model_completion=True)

    def __iter__(self):
        """Overloading of iterator from Model.  We iterate only on the
        variables defined in the assignment.
        """
        for d in self.z3_model.decls():
            if d.arity() == 0:
                try:
                    pysmt_d = self.converter.back(d())
                    yield pysmt_d, self.get_value(pysmt_d)
                except UndefinedSymbolError:
                    # avoids problems with symbols generated by z3
                    pass

    def __contains__(self, x):
        """Returns whether the model contains a value for 'x'."""
        return x in (v for v, _ in self)


# EOC Z3Model


class Z3Options(SolverOptions):
    @staticmethod
    def _set_option(z3solver, name, value):
        try:
            z3solver.set(name, value)
        except z3.Z3Exception:
            raise PysmtValueError("Error setting the option '%s=%s'" % (name, value))
        except z3.z3types.Z3Exception:
            raise PysmtValueError("Error setting the option '%s=%s'" % (name, value))

    def __call__(self, solver):
        self._set_option(solver.z3, "model", self.generate_models)
        if self.unsat_cores_mode is not None:
            self._set_option(solver.z3, "unsat_core", True)
        if self.random_seed is not None:
            self._set_option(solver.z3, "random_seed", self.random_seed)
        for k, v in self.solver_options.items():
            try:
                self._set_option(solver.z3, str(k), v)
            except z3.Z3Exception:
                raise PysmtValueError("Error setting the option '%s=%s'" % (k, v))
            except z3.z3types.Z3Exception:
                raise PysmtValueError("Error setting the option '%s=%s'" % (k, v))


# EOC Z3Options


class Z3Solver(
    IncrementalTrackingSolver, UnsatCoreSolver, SmtLibBasicSolver, SmtLibIgnoreMixin
):

    LOGICS = PYSMT_LOGICS - set(x for x in PYSMT_LOGICS if x.theory.strings)
    OptionsClass = Z3Options

    SOLVERFOR_LOGIC_NAMES = [
        "AUFLIA",
        "ALIA",
        "AUFLIRA",
        "AUFNIRA",
        "LRA",
        "LIA",
        "NIA",
        "NRA",
        "QF_ABV",
        "QF_AUFBV",
        "QF_AUFLIA",
        "QF_ALIA",
        "QF_AX",
        "QF_BV",
        "BV",
        "UFBV",
        "QF_IDL",
        "QF_LIA",
        "QF_LRA",
        "QF_NIA",
        "QF_NRA",
        "QF_RDL",
        "QF_UF",
        "UF",
        "QF_UFBV",
        "QF_UFIDL",
        "QF_UFLIA",
        "QF_UFLRA",
        "QF_UFNRA",
        "QF_UFNIA",
        "UFLRA",
        "UFNIA",
    ]

    def __init__(self, environment, logic, **options):
        IncrementalTrackingSolver.__init__(
            self, environment=environment, logic=logic, **options
        )
        # LBYL to avoid a possible segmentation fault caused by z3.SolverFor
        # See issue #465 (https://github.com/pysmt/pysmt/issues/465)
        if str(logic) in Z3Solver.SOLVERFOR_LOGIC_NAMES:
            self.z3 = z3.SolverFor(str(logic))
        else:
            self.z3 = z3.Solver()
        self.options(self)
        self.declarations = set()
        self.converter = Z3Converter(environment, z3_ctx=self.z3.ctx)
        self.mgr = environment.formula_manager

        self._name_cnt = 0
        return

    @clear_pending_pop
    def _reset_assertions(self):
        self.z3.reset()
        self.options(self)

    @clear_pending_pop
    def declare_variable(self, var):
        raise NotImplementedError

    @clear_pending_pop
    def _add_assertion(self, formula, named=None):
        self._assert_is_boolean(formula)
        term = self.converter.convert(formula)

        if self.options.unsat_cores_mode is not None:
            # TODO: IF unsat_cores_mode is all, then we add this fresh variable.
            # Otherwise, we should track this only if it is named.
            key = self.mgr.FreshSymbol(template="_assertion_%d")
            tkey = self.converter.convert(key)
            self.z3.assert_and_track(term, tkey)
            return (key, named, formula)
        else:
            self.z3.add(term)
            return formula

    def get_model(self):
        return Z3Model(self.environment, self.z3.model())

    @clear_pending_pop
    def _solve(self, assumptions=None):
        if assumptions is not None:
            bool_ass = []
            other_ass = []
            for x in assumptions:
                if x.is_literal():
                    bool_ass.append(self.converter.convert(x))
                else:
                    other_ass.append(x)

            if len(other_ass) > 0:
                self.push()
                self.add_assertion(self.mgr.And(other_ass))
                self.pending_pop = True
            res = self.z3.check(*bool_ass)
        else:
            res = self.z3.check()

        sres = str(res)
        assert sres in ["unknown", "sat", "unsat"]
        if sres == "unknown":
            raise SolverReturnedUnknownResultError
        return sres == "sat"

    def get_unsat_core(self):
        """After a call to solve() yielding UNSAT, returns the unsat core as a
        set of formulae"""
        return self.get_named_unsat_core().values()

    def _named_assertions_map(self):
        if self.options.unsat_cores_mode is not None:
            return dict((t[0], (t[1], t[2])) for t in self.assertions)
        return None

    def get_named_unsat_core(self):
        """After a call to solve() yielding UNSAT, returns the unsat core as a
        dict of names to formulae"""
        if self.options.unsat_cores_mode is None:
            raise SolverNotConfiguredForUnsatCoresError

        if self.last_result is not False:
            raise SolverStatusError("The last call to solve() was not" " unsatisfiable")

        if self.last_command != "solve":
            raise SolverStatusError(
                "The solver status has been modified by a"
                " '%s' command after the last call to"
                " solve()" % self.last_command
            )

        assumptions = self.z3.unsat_core()
        pysmt_assumptions = set(self.converter.back(t) for t in assumptions)

        res = {}
        n_ass_map = self._named_assertions_map()
        cnt = 0
        for key in pysmt_assumptions:
            if key in n_ass_map:
                (name, formula) = n_ass_map[key]
                if name is None:
                    name = "_a_%d" % cnt
                    cnt += 1
                res[name] = formula
        return res

    @clear_pending_pop
    def all_sat(self, important, callback):
        raise NotImplementedError

    @clear_pending_pop
    def _push(self, levels=1):
        for _ in range(levels):
            self.z3.push()

    @clear_pending_pop
    def _pop(self, levels=1):
        for _ in range(levels):
            self.z3.pop()

    def print_model(self, name_filter=None):
        for var in self.declarations:
            if name_filter is None or not var.symbol_name().startswith(name_filter):
                print("%s = %s" % (var.symbol_name(), self.get_value(var)))

    def get_value(self, item):
        self._assert_no_function_type(item)

        titem = self.converter.convert(item)
        z3_res = self.z3.model().eval(titem, model_completion=True)
        res = self.converter.back(z3_res, self.z3.model())
        if not res.is_constant():
            return res.simplify()
        return res

    def _exit(self):
        del self.converter
        del self.z3


BOOLREF_SET = op.BOOL_OPERATORS | op.RELATIONS
ARITHREF_SET = op.IRA_OPERATORS
BITVECREF_SET = op.BV_OPERATORS


class Z3Converter(Converter, DagWalker):
    def __init__(self, environment, z3_ctx):
        DagWalker.__init__(self, environment)
        self.mgr = environment.formula_manager
        self._get_type = environment.stc.get_type
        self._back_memoization = {}
        self.ctx = z3_ctx

        # Back Conversion
        self._back_fun = {
            z3.Z3_OP_AND: lambda args, expr: self.mgr.And(args),
            z3.Z3_OP_OR: lambda args, expr: self.mgr.Or(args),
            z3.Z3_OP_MUL: lambda args, expr: self.mgr.Times(args),
            z3.Z3_OP_ADD: lambda args, expr: self.mgr.Plus(args),
            z3.Z3_OP_DIV: lambda args, expr: self.mgr.Div(args[0], args[1]),
            z3.Z3_OP_IFF: lambda args, expr: self.mgr.Iff(args[0], args[1]),
            z3.Z3_OP_XOR: lambda args, expr: self.mgr.Xor(args[0], args[1]),
            z3.Z3_OP_FALSE: lambda args, expr: self.mgr.FALSE(),
            z3.Z3_OP_TRUE: lambda args, expr: self.mgr.TRUE(),
            z3.Z3_OP_GT: lambda args, expr: self.mgr.GT(args[0], args[1]),
            z3.Z3_OP_GE: lambda args, expr: self.mgr.GE(args[0], args[1]),
            z3.Z3_OP_LT: lambda args, expr: self.mgr.LT(args[0], args[1]),
            z3.Z3_OP_LE: lambda args, expr: self.mgr.LE(args[0], args[1]),
            z3.Z3_OP_SUB: lambda args, expr: self.mgr.Minus(args[0], args[1]),
            z3.Z3_OP_NOT: lambda args, expr: self.mgr.Not(args[0]),
            z3.Z3_OP_IMPLIES: lambda args, expr: self.mgr.Implies(args[0], args[1]),
            z3.Z3_OP_ITE: lambda args, expr: self.mgr.Ite(args[0], args[1], args[2]),
            z3.Z3_OP_TO_REAL: lambda args, expr: self.mgr.ToReal(args[0]),
            z3.Z3_OP_BAND: lambda args, expr: self.mgr.BVAnd(args[0], args[1]),
            z3.Z3_OP_BOR: lambda args, expr: self.mgr.BVOr(args[0], args[1]),
            z3.Z3_OP_BXOR: lambda args, expr: self.mgr.BVXor(args[0], args[1]),
            z3.Z3_OP_BNOT: lambda args, expr: self.mgr.BVNot(args[0]),
            z3.Z3_OP_BNEG: lambda args, expr: self.mgr.BVNeg(args[0]),
            z3.Z3_OP_CONCAT: lambda args, expr: self.mgr.BVConcat(args[0], args[1]),
            z3.Z3_OP_ULT: lambda args, expr: self.mgr.BVULT(args[0], args[1]),
            z3.Z3_OP_ULEQ: lambda args, expr: self.mgr.BVULE(args[0], args[1]),
            z3.Z3_OP_SLT: lambda args, expr: self.mgr.BVSLT(args[0], args[1]),
            z3.Z3_OP_SLEQ: lambda args, expr: self.mgr.BVSLE(args[0], args[1]),
            z3.Z3_OP_UGT: lambda args, expr: self.mgr.BVUGT(args[0], args[1]),
            z3.Z3_OP_UGEQ: lambda args, expr: self.mgr.BVUGE(args[0], args[1]),
            z3.Z3_OP_SGT: lambda args, expr: self.mgr.BVSGT(args[0], args[1]),
            z3.Z3_OP_SGEQ: lambda args, expr: self.mgr.BVSGE(args[0], args[1]),
            z3.Z3_OP_BADD: lambda args, expr: self.mgr.BVAdd(args[0], args[1]),
            z3.Z3_OP_BMUL: lambda args, expr: self.mgr.BVMul(args[0], args[1]),
            z3.Z3_OP_BUDIV: lambda args, expr: self.mgr.BVUDiv(args[0], args[1]),
            z3.Z3_OP_BSDIV: lambda args, expr: self.mgr.BVSDiv(args[0], args[1]),
            z3.Z3_OP_BUREM: lambda args, expr: self.mgr.BVURem(args[0], args[1]),
            z3.Z3_OP_BSREM: lambda args, expr: self.mgr.BVSRem(args[0], args[1]),
            z3.Z3_OP_BSHL: lambda args, expr: self.mgr.BVLShl(args[0], args[1]),
            z3.Z3_OP_BLSHR: lambda args, expr: self.mgr.BVLShr(args[0], args[1]),
            z3.Z3_OP_BASHR: lambda args, expr: self.mgr.BVAShr(args[0], args[1]),
            z3.Z3_OP_BSUB: lambda args, expr: self.mgr.BVSub(args[0], args[1]),
            z3.Z3_OP_EXT_ROTATE_LEFT: lambda args, expr: self.mgr.BVRol(
                args[0], args[1].bv_unsigned_value()
            ),
            z3.Z3_OP_EXT_ROTATE_RIGHT: lambda args, expr: self.mgr.BVRor(
                args[0], args[1].bv_unsigned_value()
            ),
            z3.Z3_OP_BV2INT: lambda args, expr: self.mgr.BVToNatural(args[0]),
            z3.Z3_OP_POWER: lambda args, expr: self.mgr.Pow(args[0], args[1]),
            z3.Z3_OP_SELECT: lambda args, expr: self.mgr.Select(args[0], args[1]),
            z3.Z3_OP_STORE: lambda args, expr: self.mgr.Store(
                args[0], args[1], args[2]
            ),
            # Actually use both args, expr
            z3.Z3_OP_SIGN_EXT: lambda args, expr: self.mgr.BVSExt(
                args[0], z3.get_payload(expr, 0)
            ),
            z3.Z3_OP_ZERO_EXT: lambda args, expr: self.mgr.BVZExt(
                args[0], z3.get_payload(expr, 0)
            ),
            z3.Z3_OP_ROTATE_LEFT: lambda args, expr: self.mgr.BVRol(
                args[0], z3.get_payload(expr, 0)
            ),
            z3.Z3_OP_ROTATE_RIGHT: lambda args, expr: self.mgr.BVRor(
                args[0], z3.get_payload(expr, 0)
            ),
            z3.Z3_OP_EXTRACT: lambda args, expr: self.mgr.BVExtract(
                args[0], z3.get_payload(expr, 1), z3.get_payload(expr, 0)
            ),
            # Complex Back Translation
            z3.Z3_OP_EQ: self._back_z3_eq,
            z3.Z3_OP_UMINUS: self._back_z3_uminus,
            z3.Z3_OP_CONST_ARRAY: self._back_z3_const_array,
        }
        # Unique reference to Sorts
        self.z3RealSort = z3.RealSort(self.ctx)
        self.z3BoolSort = z3.BoolSort(self.ctx)
        self.z3IntSort = z3.IntSort(self.ctx)
        self._z3ArraySorts = {}
        self._z3BitVecSorts = {}
        self._z3Sorts = {}
        # Unique reference to Function Declaration
        self._z3_func_decl_cache = {}
        return

    def z3BitVecSort(self, width):
        """Return the z3 BitVecSort for the given width."""
        try:
            bvsort = self._z3BitVecSorts[width]
        except KeyError:
            bvsort = z3.BitVecSort(width, self.ctx)
            self._z3BitVecSorts[width] = bvsort
        return bvsort

    def z3ArraySort(self, key, value):
        """Return the z3 ArraySort for the given key value."""
        try:
            return self._z3ArraySorts[(askey(key), askey(value))]
        except KeyError:
            sort = z3.ArraySort(key, value)
            self._z3ArraySorts[(askey(key), askey(value))] = sort
        return sort

    def z3Sort(self, name):
        """Return the z3 Sort for the given name."""
        name = str(name)
        try:
            return self._z3Sorts[name]
        except KeyError:
            sort = z3.DeclareSort(name, self.ctx)
            self._z3Sorts[name] = sort
        return sort

    def get_z3_ref(self, formula):
        if formula.node_type in op.QUANTIFIERS:
            return z3.QuantifierRef
        elif formula.node_type() in BOOLREF_SET:
            return z3.BoolRef
        elif formula.node_type() in ARITHREF_SET:
            return z3.ArithRef
        elif formula.node_type() in BITVECREF_SET:
            return z3.BitVecRef
        elif formula.is_symbol() or formula.is_function_application():
            if formula.is_function_application():
                type_ = formula.function_name().symbol_type()
                type_ = type_.return_type
            else:
                type_ = formula.symbol_type()

            if type_.is_bool_type():
                return z3.BoolRef
            elif type_.is_real_type() or type_.is_int_type():
                return z3.ArithRef
            elif type_.is_array_type():
                return z3.ArrayRef
            elif type_.is_bv_type():
                return z3.BitVecRef
            else:
                raise NotImplementedError(formula)
        elif formula.node_type() in op.ARRAY_OPERATORS:
            return z3.ArrayRef
        elif formula.is_ite():
            child = formula.arg(1)
            return self.get_z3_ref(child)
        else:
            assert formula.is_constant(), formula
            type_ = formula.constant_type()
            if type_.is_bool_type():
                return z3.BoolRef
            elif type_.is_real_type() or type_.is_int_type():
                return z3.ArithRef
            elif type_.is_array_type():
                return z3.ArrayRef
            elif type_.is_bv_type():
                return z3.BitVecRef
            else:
                raise NotImplementedError(formula)

    @catch_conversion_error
    def convert(self, formula):
        z3term = self.walk(formula)
        ref_class = self.get_z3_ref(formula)
        return ref_class(z3term, self.ctx)

    def back(self, expr, model=None):
        """Convert a Z3 expression back into a pySMT expression.

        This is done using the Z3 API. For very big expressions, it is
        sometimes faster to go through the SMT-LIB format. In those
        cases, consider using the method back_via_smtlib.
        """
        stack = [expr]
        while len(stack) > 0:
            current = stack.pop()
            key = (askey(current), model)
            if key not in self._back_memoization:
                self._back_memoization[key] = None
                stack.append(current)
                for c in current.children():
                    stack.append(c)
            elif self._back_memoization[key] is None:
                args = [
                    self._back_memoization[(askey(c), model)]
                    for c in current.children()
                ]
                res = self._back_single_term(current, args, model)
                self._back_memoization[key] = res
            else:
                # we already visited the node, nothing else to do
                pass
        return self._back_memoization[(askey(expr), model)]

    def _back_single_term(self, expr, args, model=None):
        assert z3.is_expr(expr)

        if z3.is_quantifier(expr):
            raise NotImplementedError(
                "Quantified back conversion is currently not supported"
            )

        assert not len(args) > 2 or (
            z3.is_and(expr)
            or z3.is_or(expr)
            or z3.is_add(expr)
            or z3.is_mul(expr)
            or (len(args) == 3 and (z3.is_ite(expr) or z3.is_array_store(expr)))
        ), ("Unexpected n-ary term: %s" % expr)

        res = None
        try:
            decl = z3.Z3_get_app_decl(expr.ctx_ref(), expr.as_ast())
            kind = z3.Z3_get_decl_kind(expr.ctx.ref(), decl)
            # Try to get the back-conversion function for the given Kind
            fun = self._back_fun[kind]
            return fun(args, expr)
        except KeyError as ex:
            pass

        if z3.is_const(expr):
            # Const or Symbol
            if z3.is_rational_value(expr):
                n = expr.numerator_as_long()
                d = expr.denominator_as_long()
                f = Fraction(n, d)
                return self.mgr.Real(f)
            elif z3.is_int_value(expr):
                n = expr.as_long()
                return self.mgr.Int(n)
            elif z3.is_bv_value(expr):
                n = expr.as_long()
                w = expr.size()
                return self.mgr.BV(n, w)
            elif z3.is_as_array(expr):
                if model is None:
                    raise NotImplementedError(
                        "As-array expressions cannot be"
                        " handled as they are not "
                        "self-contained"
                    )
                else:
                    interp_decl = z3.get_as_array_func(expr)
                    interp = model[interp_decl]
                    default = self.back(interp.else_value(), model=model)
                    assign = {}
                    for i in range(interp.num_entries()):
                        e = interp.entry(i)
                        assert e.num_args() == 1
                        idx = self.back(e.arg_value(0), model=model)
                        val = self.back(e.value(), model=model)
                        assign[idx] = val
                    arr_type = self._z3_to_type(expr.sort())
                    return self.mgr.Array(arr_type.index_type, default, assign)
            elif z3.is_algebraic_value(expr):
                # Algebraic value
                return self.mgr._Algebraic(Numeral(expr))
            else:
                # it must be a symbol
                try:
                    return self.mgr.get_symbol(str(expr))
                except UndefinedSymbolError:
                    import warnings

                    symb_type = self._z3_to_type(expr.sort())
                    warnings.warn("Defining new symbol: %s" % str(expr))
                    return self.mgr.FreshSymbol(symb_type, template="__z3_%d")
        elif z3.is_function(expr):
            # This needs to be after we try to convert regular Symbols
            fsymbol = self.mgr.get_symbol(expr.decl().name())
            return self.mgr.Function(fsymbol, args)

        # If we reach this point, we did not manage to translate the expression
        raise ConvertExpressionError(
            message=("Unsupported expression: %s" % (str(expr))), expression=expr
        )

    def _back_z3_eq(self, args, expr):
        if self._get_type(args[0]).is_bool_type():
            return self.mgr.Iff(args[0], args[1])
        return self.mgr.Equals(args[0], args[1])

    def _back_z3_uminus(self, args, expr):
        tp = self._get_type(args[0])
        if tp.is_real_type():
            minus_one = self.mgr.Real(-1)
        else:
            assert tp.is_int_type()
            minus_one = self.mgr.Int(-1)
        return self.mgr.Times(args[0], minus_one)

    def _back_z3_const_array(self, args, expr):
        arr_ty = self._z3_to_type(expr.sort())
        return self.mgr.Array(arr_ty.index_type, args[0])

    def back_via_smtlib(self, expr):
        """Back convert a Z3 Expression by translation to SMT-LIB."""
        from io import StringIO
        from pysmt.smtlib.parser import SmtLibZ3Parser

        parser = SmtLibZ3Parser(self.env)

        z3.Z3_set_ast_print_mode(expr.ctx.ref(), z3.Z3_PRINT_SMTLIB2_COMPLIANT)
        s = z3.Z3_benchmark_to_smtlib_string(
            expr.ctx.ref(), None, None, None, None, 0, None, expr.ast
        )
        stream_in = StringIO(s)
        r = parser.get_script(stream_in).get_last_formula(self.mgr)
        key = (askey(expr), None)
        self._back_memoization[key] = r
        return r

    # Fwd Conversion
    def _to_ast_array(self, args):
        """Convert a list of arguments into an z3.AST vector."""
        sz = len(args)
        _args = (z3.Ast * sz)()
        for i, arg in enumerate(args):
            _args[i] = arg
        return _args, sz

    def walk_not(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_not(self.ctx.ref(), args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_symbol(self, formula, **kwargs):
        symbol_type = formula.symbol_type()
        sname = formula.symbol_name()
        z3_sname = z3.Z3_mk_string_symbol(self.ctx.ref(), sname)
        if symbol_type.is_bool_type():
            sort_ast = self.z3BoolSort.ast
        elif symbol_type.is_real_type():
            sort_ast = self.z3RealSort.ast
        elif symbol_type.is_int_type():
            sort_ast = self.z3IntSort.ast
        elif symbol_type.is_array_type():
            sort_ast = self._type_to_z3(symbol_type).ast
        elif symbol_type.is_string_type():
            raise ConvertExpressionError(
                message=("Unsupported string symbol: %s" % str(formula)),
                expression=formula,
            )
        else:
            sort_ast = self._type_to_z3(symbol_type).ast
        # Create const with given sort
        res = z3.Z3_mk_const(self.ctx.ref(), z3_sname, sort_ast)
        z3.Z3_inc_ref(self.ctx.ref(), res)
        return res

    def walk_ite(self, formula, args, **kwargs):
        i = args[0]
        ni = self.walk_not(None, (i,))
        t = args[1]
        e = args[2]

        if self._get_type(formula).is_bool_type():
            # Rewrite as (!i \/ t) & (i \/ e)
            _args, sz = self._to_ast_array((ni, t))
            or1 = z3.Z3_mk_or(self.ctx.ref(), sz, _args)
            z3.Z3_inc_ref(self.ctx.ref(), or1)
            _args, sz = self._to_ast_array((i, e))
            or2 = z3.Z3_mk_or(self.ctx.ref(), sz, _args)
            z3.Z3_inc_ref(self.ctx.ref(), or2)
            _args, sz = self._to_ast_array((or1, or2))
            z3term = z3.Z3_mk_and(self.ctx.ref(), sz, _args)
            z3.Z3_inc_ref(self.ctx.ref(), z3term)
            z3.Z3_dec_ref(self.ctx.ref(), or1)
            z3.Z3_dec_ref(self.ctx.ref(), or2)
            return z3term
        z3term = z3.Z3_mk_ite(self.ctx.ref(), i, t, e)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_algebraic_constant(self, formula, **kwargs):
        rep = str(formula.constant_value())
        z3term = z3.Z3_mk_numeral(self.ctx.ref(), rep, self.z3RealSort.ast)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_real_constant(self, formula, **kwargs):
        frac = formula.constant_value()
        n, d = frac.numerator, frac.denominator
        rep = str(n) + "/" + str(d)
        z3term = z3.Z3_mk_numeral(self.ctx.ref(), rep, self.z3RealSort.ast)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_int_constant(self, formula, **kwargs):
        assert is_pysmt_integer(formula.constant_value())
        const = str(formula.constant_value())
        z3term = z3.Z3_mk_numeral(self.ctx.ref(), const, self.z3IntSort.ast)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bool_constant(self, formula, **kwargs):
        _t = z3.BoolVal(formula.constant_value(), ctx=self.ctx)
        z3term = _t.as_ast()
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_quantifier(self, formula, args, **kwargs):
        qvars = formula.quantifier_vars()
        qvars, qvars_sz = self._to_ast_array([self.walk_symbol(x) for x in qvars])
        empty_str = z3.Z3_mk_string_symbol(self.ctx.ref(), "")
        z3term = z3.Z3_mk_quantifier_const_ex(
            self.ctx.ref(),
            formula.is_forall(),
            1,
            empty_str,
            empty_str,
            qvars_sz,
            qvars,
            0,
            None,
            0,
            None,
            args[0],
        )
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_toreal(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_int2real(self.ctx.ref(), args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def _z3_func_decl(self, func_name):
        """Create a Z3 Function Declaration for the given function."""
        try:
            return self._z3_func_decl_cache[func_name]
        except KeyError:
            tp = func_name.symbol_type()
            arity = len(tp.param_types)
            z3dom = (z3.Sort * arity)()
            for i, t in enumerate(tp.param_types):
                z3dom[i] = self._type_to_z3(t).ast
            z3ret = self._type_to_z3(tp.return_type).ast
            z3name = z3.Z3_mk_string_symbol(self.ctx.ref(), func_name.symbol_name())
            z3func = z3.Z3_mk_func_decl(self.ctx.ref(), z3name, arity, z3dom, z3ret)
            self._z3_func_decl_cache[func_name] = z3func
            return z3func

    def walk_function(self, formula, args, **kwargs):
        z3func = self._z3_func_decl(formula.function_name())
        _args, sz = self._to_ast_array(args)
        z3term = z3.Z3_mk_app(self.ctx.ref(), z3func, sz, _args)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_constant(self, formula, **kwargs):
        value = formula.constant_value()
        z3term = z3.Z3_mk_numeral(
            self.ctx.ref(), str(value), self.z3BitVecSort(formula.bv_width()).ast
        )
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_extract(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_extract(
            self.ctx.ref(),
            formula.bv_extract_end(),
            formula.bv_extract_start(),
            args[0],
        )
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_not(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_bvnot(self.ctx.ref(), args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_neg(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_bvneg(self.ctx.ref(), args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_rol(self, formula, args, **kwargs):
        bvsort = self.z3BitVecSort(formula.bv_width())
        step = z3.Z3_mk_numeral(
            self.ctx.ref(), str(formula.bv_rotation_step()), bvsort.ast
        )
        z3term = z3.Z3_mk_ext_rotate_left(self.ctx.ref(), args[0], step)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_ror(self, formula, args, **kwargs):
        bvsort = self.z3BitVecSort(formula.bv_width())
        step = z3.Z3_mk_numeral(
            self.ctx.ref(), str(formula.bv_rotation_step()), bvsort.ast
        )
        z3term = z3.Z3_mk_ext_rotate_right(self.ctx.ref(), args[0], step)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_zext(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_zero_ext(self.ctx.ref(), formula.bv_extend_step(), args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_sext(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_sign_ext(self.ctx.ref(), formula.bv_extend_step(), args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_bv_comp(self, formula, args, **kwargs):
        cond = z3.Z3_mk_eq(self.ctx.ref(), args[0], args[1])
        z3.Z3_inc_ref(self.ctx.ref(), cond)
        then_ = z3.Z3_mk_numeral(self.ctx.ref(), "1", self.z3BitVecSort(1).ast)
        z3.Z3_inc_ref(self.ctx.ref(), then_)
        else_ = z3.Z3_mk_numeral(self.ctx.ref(), "0", self.z3BitVecSort(1).ast)
        z3.Z3_inc_ref(self.ctx.ref(), else_)
        z3term = z3.Z3_mk_ite(self.ctx.ref(), cond, then_, else_)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        # De-Ref since this is handled internally by Z3
        z3.Z3_dec_ref(self.ctx.ref(), cond)
        z3.Z3_dec_ref(self.ctx.ref(), then_)
        z3.Z3_dec_ref(self.ctx.ref(), else_)
        return z3term

    def walk_bv_tonatural(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_bv2int(self.ctx.ref(), args[0], False)
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_array_select(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_select(self.ctx.ref(), args[0], args[1])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_array_store(self, formula, args, **kwargs):
        z3term = z3.Z3_mk_store(self.ctx.ref(), args[0], args[1], args[2])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def walk_array_value(self, formula, args, **kwargs):
        idx_type = formula.array_value_index_type()
        arraysort = self._type_to_z3(idx_type).ast
        z3term = z3.Z3_mk_const_array(self.ctx.ref(), arraysort, args[0])
        z3.Z3_inc_ref(self.ctx.ref(), z3term)

        for i in range(1, len(args), 2):
            c = args[i]
            z3term = self.walk_array_store(None, (z3term, c, args[i + 1]))
            z3.Z3_inc_ref(self.ctx.ref(), z3term)
        return z3term

    def _z3_to_type(self, sort):
        if sort.kind() == z3.Z3_BOOL_SORT:
            return types.BOOL
        elif sort.kind() == z3.Z3_INT_SORT:
            return types.INT
        elif sort.kind() == z3.Z3_REAL_SORT:
            return types.REAL
        elif sort.kind() == z3.Z3_ARRAY_SORT:
            return types.ArrayType(
                self._z3_to_type(sort.domain()), self._z3_to_type(sort.range())
            )
        elif sort.kind() == z3.Z3_BV_SORT:
            return types.BVType(sort.size())
        else:
            raise NotImplementedError("Unsupported sort in conversion: %s" % sort)

    def make_walk_nary(func):
        def walk_nary(self, formula, args, **kwargs):
            _args, sz = self._to_ast_array(args)
            z3term = func(self.ctx.ref(), sz, _args)
            z3.Z3_inc_ref(self.ctx.ref(), z3term)
            return z3term

        return walk_nary

    def make_walk_binary(func):
        def walk_binary(self, formula, args, **kwargs):
            z3term = func(self.ctx.ref(), args[0], args[1])
            z3.Z3_inc_ref(self.ctx.ref(), z3term)
            return z3term

        return walk_binary

    walk_and = make_walk_nary(z3.Z3_mk_and)
    walk_or = make_walk_nary(z3.Z3_mk_or)
    walk_plus = make_walk_nary(z3.Z3_mk_add)
    walk_times = make_walk_nary(z3.Z3_mk_mul)
    walk_minus = make_walk_nary(z3.Z3_mk_sub)
    walk_implies = make_walk_binary(z3.Z3_mk_implies)
    walk_le = make_walk_binary(z3.Z3_mk_le)
    walk_lt = make_walk_binary(z3.Z3_mk_lt)
    walk_equals = make_walk_binary(z3.Z3_mk_eq)
    walk_iff = make_walk_binary(z3.Z3_mk_eq)
    walk_pow = make_walk_binary(z3.Z3_mk_power)
    walk_div = make_walk_binary(z3.Z3_mk_div)
    walk_bv_ult = make_walk_binary(z3.Z3_mk_bvult)
    walk_bv_ule = make_walk_binary(z3.Z3_mk_bvule)
    walk_bv_slt = make_walk_binary(z3.Z3_mk_bvslt)
    walk_bv_sle = make_walk_binary(z3.Z3_mk_bvsle)
    walk_bv_concat = make_walk_binary(z3.Z3_mk_concat)
    walk_bv_or = make_walk_binary(z3.Z3_mk_bvor)
    walk_bv_and = make_walk_binary(z3.Z3_mk_bvand)
    walk_bv_xor = make_walk_binary(z3.Z3_mk_bvxor)
    walk_bv_add = make_walk_binary(z3.Z3_mk_bvadd)
    walk_bv_sub = make_walk_binary(z3.Z3_mk_bvsub)
    walk_bv_mul = make_walk_binary(z3.Z3_mk_bvmul)
    walk_bv_udiv = make_walk_binary(z3.Z3_mk_bvudiv)
    walk_bv_urem = make_walk_binary(z3.Z3_mk_bvurem)
    walk_bv_lshl = make_walk_binary(z3.Z3_mk_bvshl)
    walk_bv_lshr = make_walk_binary(z3.Z3_mk_bvlshr)
    walk_bv_sdiv = make_walk_binary(z3.Z3_mk_bvsdiv)
    walk_bv_srem = make_walk_binary(z3.Z3_mk_bvsrem)
    walk_bv_ashr = make_walk_binary(z3.Z3_mk_bvashr)
    walk_exists = walk_quantifier
    walk_forall = walk_quantifier

    def _type_to_z3(self, tp):
        """Convert a pySMT type into the corresponding Z3 sort."""
        if tp.is_bool_type():
            return self.z3BoolSort
        elif tp.is_real_type():
            return self.z3RealSort
        elif tp.is_int_type():
            return self.z3IntSort
        elif tp.is_array_type():
            key_sort = self._type_to_z3(tp.index_type)
            val_sort = self._type_to_z3(tp.elem_type)
            return self.z3ArraySort(key_sort, val_sort)
        elif tp.is_bv_type():
            return self.z3BitVecSort(tp.width)
        else:
            assert tp.is_custom_type(), "Unsupported type '%s'" % tp
            return self.z3Sort(tp)
        raise NotImplementedError("Unsupported type in conversion: %s" % tp)

    def __del__(self):
        # Cleaning-up Z3Converter requires dec-ref'ing the terms in the cache
        if self.ctx.ref():
            # Check that there is still a context object
            # This might not be the case if we are using the global context
            # and the interpreter is shutting down
            for t in self.memoization.values():
                z3.Z3_dec_ref(self.ctx.ref(), t)


# EOC Z3Converter


class Z3QuantifierEliminator(QuantifierEliminator):

    LOGICS = [LIA, LRA]

    def __init__(self, environment, logic=None):
        QuantifierEliminator.__init__(self)
        self.environment = environment
        self.logic = logic
        self.converter = Z3Converter(environment, z3.main_ctx())

    def eliminate_quantifiers(self, formula):
        logic = get_logic(formula, self.environment)
        if not logic <= LRA and not logic <= LIA:
            raise PysmtValueError(
                "Z3 quantifier elimination only "
                "supports LRA or LIA without combination."
                "(detected logic is: %s)" % str(logic)
            )

        simplifier = z3.Tactic("simplify")
        eliminator = z3.Tactic("qe")

        f = self.converter.convert(formula)
        s = simplifier(
            f, elim_and=True, pull_cheap_ite=True, ite_extra_rules=True
        ).as_expr()
        res = eliminator(f).as_expr()

        pysmt_res = None
        try:
            pysmt_res = self.converter.back(res)
        except ConvertExpressionError:
            if logic <= LRA:
                raise
            raise ConvertExpressionError(
                message=(
                    "Unable to represent"
                    "expression %s in pySMT: the quantifier elimination for "
                    "LIA is incomplete as it requires the modulus. You can "
                    "find the Z3 expression representing the quantifier "
                    "elimination as the attribute 'expression' of this "
                    "exception object" % str(res)
                ),
                expression=res,
            )

        return pysmt_res

    def _exit(self):
        pass



#
# This file is part of pySMT.
#
#   Copyright 2014 Andrea Micheli and Marco Gario
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import absolute_import

from pysmt.exceptions import SolverAPINotFound

try:
    import CVC4
except ImportError:
    raise SolverAPINotFound

import pysmt.typing as types
from pysmt.logics import PYSMT_LOGICS, ARRAYS_CONST_LOGICS

from pysmt.solvers.solver import Solver, Converter, SolverOptions
from pysmt.exceptions import (
    SolverReturnedUnknownResultError,
    InternalSolverError,
    NonLinearError,
    PysmtValueError,
    PysmtTypeError,
)
from pysmt.walkers import DagWalker
from pysmt.solvers.smtlib import SmtLibBasicSolver, SmtLibIgnoreMixin
from pysmt.solvers.eager import EagerModel
from pysmt.decorators import catch_conversion_error
from pysmt.constants import Fraction, is_pysmt_integer, to_python_integer


class CVC4Options(SolverOptions):
    def __init__(self, **base_options):
        SolverOptions.__init__(self, **base_options)
        # TODO: CVC4 Supports UnsatCore extraction
        # but we did not wrapped it yet. (See #349)
        if self.unsat_cores_mode is not None:
            raise PysmtValueError("'unsat_cores_mode' option not supported.")

    @staticmethod
    def _set_option(cvc4, name, value):
        try:
            cvc4.setOption(name, CVC4.SExpr(value))
        except:
            raise PysmtValueError("Error setting the option '%s=%s'" % (name, value))

    def __call__(self, solver):
        if solver.logic_name == "QF_SLIA":
            self._set_option(solver.cvc4, "strings-exp", "true")

        self._set_option(
            solver.cvc4, "produce-models", str(self.generate_models).lower()
        )
        self._set_option(solver.cvc4, "incremental", str(self.incremental).lower())
        if self.random_seed is not None:
            self._set_option(solver.cvc4, "random-seed", str(self.random_seed))

        for k, v in self.solver_options.items():
            self._set_option(solver.cvc4, str(k), str(v))


# EOC CVC4Options


class CVC4Solver(Solver, SmtLibBasicSolver, SmtLibIgnoreMixin):

    LOGICS = (
        PYSMT_LOGICS
        - ARRAYS_CONST_LOGICS
        - set(l for l in PYSMT_LOGICS if not l.theory.linear)
    )

    OptionsClass = CVC4Options

    def __init__(self, environment, logic, **options):
        Solver.__init__(self, environment=environment, logic=logic, **options)
        self.em = CVC4.ExprManager()

        self.cvc4 = None
        self.declarations = None
        self.logic_name = str(logic)
        if "t" in self.logic_name:
            # Custom Type extension
            self.logic_name = self.logic_name.replace("t", "")
        if self.logic_name == "QF_BOOL":
            self.logic_name = "QF_LRA"
        elif self.logic_name == "BOOL":
            self.logic_name = "LRA"

        self.reset_assertions()
        self.converter = CVC4Converter(environment, cvc4_exprMgr=self.em)
        return

    def reset_assertions(self):
        del self.cvc4
        # CVC4's SWIG interface is not acquiring ownership of the
        # SmtEngine object. Forcing it here.
        self.cvc4 = CVC4.SmtEngine(self.em)
        self.cvc4.thisown = 1
        self.options(self)
        self.declarations = set()
        self.cvc4.setLogic(self.logic_name)

    def declare_variable(self, var):
        raise NotImplementedError

    def add_assertion(self, formula, named=None):
        self._assert_is_boolean(formula)
        term = self.converter.convert(formula)
        self.cvc4.assertFormula(term)
        return

    def get_model(self):
        assignment = {}
        for s in self.environment.formula_manager.get_all_symbols():
            if s.is_term():
                if s.symbol_type().is_custom_type():
                    continue
                v = self.get_value(s)
                assignment[s] = v
        return EagerModel(assignment=assignment, environment=self.environment)

    def solve(self, assumptions=None):
        if assumptions is not None:
            conj_assumptions = self.environment.formula_manager.And(assumptions)
            cvc4_assumption = self.converter.convert(conj_assumptions)
            res = self.cvc4.checkSat(cvc4_assumption)
        else:
            try:
                res = self.cvc4.checkSat()
            except:
                raise InternalSolverError()

        # Convert returned type
        res_type = res.isSat()
        if res_type == CVC4.Result.SAT_UNKNOWN:
            raise SolverReturnedUnknownResultError()
        else:
            return res_type == CVC4.Result.SAT
        return

    def push(self, levels=1):
        if not self.options.incremental:
            # The exceptions from CVC4 are not raised correctly
            # (probably due to the wrapper)
            # Therefore, we need to check that we can actually do a push
            raise NotImplementedError("The solver is not incremental")

        for _ in range(levels):
            self.cvc4.push()
        return

    def pop(self, levels=1):
        for _ in range(levels):
            self.cvc4.pop()
        return

    def print_model(self, name_filter=None):
        if name_filter is None:
            var_set = self.declarations
        else:
            var_set = (var for var in self.declarations if name_filter(var))
        for var in var_set:
            print("%s = %s", (var.symbol_name(), self.get_value(var)))
        return

    def get_value(self, item):
        self._assert_no_function_type(item)
        term = self.converter.convert(item)
        cvc4_res = self.cvc4.getValue(term)
        res = self.converter.back(cvc4_res)
        if (
            self.environment.stc.get_type(item).is_real_type()
            and self.environment.stc.get_type(res).is_int_type()
        ):
            res = self.environment.formula_manager.Real(
                Fraction(res.constant_value(), 1)
            )
        return res

    def _exit(self):
        del self.cvc4

    def set_option(self, name, value):
        """Sets an option.

        :param name and value: Option to be set
        :type name: String
        :type value: String
        """
        self.cvc4.setOption(name, CVC4.SExpr(value))


class CVC4Converter(Converter, DagWalker):
    def __init__(self, environment, cvc4_exprMgr):
        DagWalker.__init__(self, environment)

        self.cvc4_exprMgr = cvc4_exprMgr
        self.mkExpr = cvc4_exprMgr.mkExpr
        self.mkConst = cvc4_exprMgr.mkConst
        self.mkVar = cvc4_exprMgr.mkVar
        self.realType = cvc4_exprMgr.realType()
        self.intType = cvc4_exprMgr.integerType()
        self.boolType = cvc4_exprMgr.booleanType()
        self.stringType = cvc4_exprMgr.stringType()

        self.declared_vars = {}
        self.backconversion = {}
        self.mgr = environment.formula_manager
        self._get_type = environment.stc.get_type
        return

    def declare_variable(self, var):
        if not var.is_symbol():
            raise PysmtTypeError(
                "Trying to declare as a variable something "
                "that is not a symbol: %s" % var
            )
        if var.symbol_name() not in self.declared_vars:
            cvc4_type = self._type_to_cvc4(var.symbol_type())
            decl = self.mkVar(var.symbol_name(), cvc4_type)
            self.declared_vars[var] = decl

    def back(self, expr):
        res = None
        if expr.isConst():
            if expr.getType().isBoolean():
                v = expr.getConstBoolean()
                res = self.mgr.Bool(v)
            elif expr.getType().isInteger():
                v = expr.getConstRational().toString()
                res = self.mgr.Int(int(v))
            elif expr.getType().isReal():
                v = expr.getConstRational().toString()
                res = self.mgr.Real(Fraction(v))
            elif expr.getType().isBitVector():
                bv = expr.getConstBitVector()
                v = bv.getValue().toString()
                width = bv.getSize()
                res = self.mgr.BV(int(v), width)
            elif expr.getType().isString():
                v = expr.getConstString()
                res = self.mgr.String(v.toString())
            elif expr.getType().isArray():
                const_ = expr.getConstArrayStoreAll()
                array_type = self._cvc4_type_to_type(const_.getType())
                base_value = self.back(const_.getExpr())
                res = self.mgr.Array(array_type.index_type, base_value)
            else:
                raise PysmtTypeError(
                    "Unsupported constant type:", expr.getType().toString()
                )
        else:
            raise PysmtTypeError("Unsupported expression:", expr.toString())

        return res

    @catch_conversion_error
    def convert(self, formula):
        return self.walk(formula)

    def walk_and(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.AND, args)

    def walk_or(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.OR, args)

    def walk_not(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.NOT, args[0])

    def walk_symbol(self, formula, args, **kwargs):
        # pylint: disable=unused-argument
        if formula not in self.declared_vars:
            self.declare_variable(formula)
        return self.declared_vars[formula]

    def walk_iff(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.EQUAL, args[0], args[1])

    def walk_implies(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.IMPLIES, args[0], args[1])

    def walk_le(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.LEQ, args[0], args[1])

    def walk_lt(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.LT, args[0], args[1])

    def walk_ite(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.ITE, args[0], args[1], args[2])

    def walk_real_constant(self, formula, **kwargs):
        frac = formula.constant_value()
        n, d = frac.numerator, frac.denominator
        rep = str(n) + "/" + str(d)
        return self.mkConst(CVC4.Rational(rep))

    def walk_int_constant(self, formula, **kwargs):
        assert is_pysmt_integer(formula.constant_value())
        rep = str(formula.constant_value())
        return self.mkConst(CVC4.Rational(rep))

    def walk_bool_constant(self, formula, **kwargs):
        return self.cvc4_exprMgr.mkBoolConst(formula.constant_value())

    def walk_exists(self, formula, args, **kwargs):
        (bound_formula, var_list) = self._rename_bound_variables(
            args[0], formula.quantifier_vars()
        )
        bound_vars_list = self.mkExpr(CVC4.BOUND_VAR_LIST, var_list)
        return self.mkExpr(CVC4.EXISTS, bound_vars_list, bound_formula)

    def walk_forall(self, formula, args, **kwargs):
        (bound_formula, var_list) = self._rename_bound_variables(
            args[0], formula.quantifier_vars()
        )
        bound_vars_list = self.mkExpr(CVC4.BOUND_VAR_LIST, var_list)
        return self.mkExpr(CVC4.FORALL, bound_vars_list, bound_formula)

    def walk_plus(self, formula, args, **kwargs):
        res = self.mkExpr(CVC4.PLUS, args)
        return res

    def walk_array_store(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STORE, args[0], args[1], args[2])

    def walk_array_select(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.SELECT, args[0], args[1])

    def walk_minus(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.MINUS, args[0], args[1])

    def walk_equals(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.EQUAL, args[0], args[1])

    def walk_times(self, formula, args, **kwargs):
        if sum(1 for x in formula.args() if x.get_free_variables()) > 1:
            raise NonLinearError(formula)
        res = args[0]
        for x in args[1:]:
            res = self.mkExpr(CVC4.MULT, res, x)
        return res

    def walk_toreal(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.TO_REAL, args[0])

    def walk_function(self, formula, args, **kwargs):
        name = formula.function_name()
        if name not in self.declared_vars:
            self.declare_variable(name)
        decl = self.declared_vars[name]
        return self.mkExpr(CVC4.APPLY_UF, decl, args)

    def walk_bv_constant(self, formula, **kwargs):
        vrepr = str(formula.constant_value())
        width = formula.bv_width()
        return self.mkConst(CVC4.BitVector(width, CVC4.Integer(vrepr)))

    def walk_bv_ult(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_ULT, args[0], args[1])

    def walk_bv_ule(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_ULE, args[0], args[1])

    def walk_bv_concat(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_CONCAT, args[0], args[1])

    def walk_bv_extract(self, formula, args, **kwargs):
        ext = self.mkConst(
            CVC4.BitVectorExtract(formula.bv_extract_end(), formula.bv_extract_start())
        )
        return self.mkExpr(CVC4.BITVECTOR_EXTRACT, ext, args[0])

    def walk_bv_or(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_OR, args[0], args[1])

    def walk_bv_not(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_NOT, args[0])

    def walk_bv_and(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_AND, args[0], args[1])

    def walk_bv_xor(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_XOR, args[0], args[1])

    def walk_bv_add(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_PLUS, args[0], args[1])

    def walk_bv_sub(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_SUB, args[0], args[1])

    def walk_bv_neg(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_NEG, args[0])

    def walk_bv_mul(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_MULT, args[0], args[1])

    def walk_bv_tonatural(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_TO_NAT, args[0])

    def walk_bv_udiv(self, formula, args, **kwargs):
        # Force deterministic semantics of division by 0
        # If the denominator is bv0, then the result is ~0
        n, d = args
        if d.isConst():
            bv = d.getConstBitVector()
            v = bv.getValue().toString()
            if v == "0":
                return self.mkExpr(CVC4.BITVECTOR_NOT, d)
            else:
                return self.mkExpr(CVC4.BITVECTOR_UDIV, n, d)
        else:
            # (d == 0) ? ~0 : n bvudiv d
            base = self.mkExpr(CVC4.BITVECTOR_UDIV, n, d)
            zero = self.mkConst(CVC4.BitVector(formula.bv_width(), CVC4.Integer("0")))
            notzero = self.mkExpr(CVC4.BITVECTOR_NOT, zero)
            test = self.mkExpr(CVC4.EQUAL, d, zero)
            return self.mkExpr(CVC4.ITE, test, notzero, base)

    def walk_bv_urem(self, formula, args, **kwargs):
        # Force deterministic semantics of reminder by 0
        # If the denominator is bv0, then the result is the numerator
        n, d = args
        if d.isConst():
            bv = d.getConstBitVector()
            v = bv.getValue().toString()
            if v == "0":
                return n
            else:
                return self.mkExpr(CVC4.BITVECTOR_UREM, n, d)
        else:
            # (d == 0) ? n : n bvurem d
            base = self.mkExpr(CVC4.BITVECTOR_UREM, n, d)
            zero = self.mkConst(CVC4.BitVector(formula.bv_width(), CVC4.Integer("0")))
            test = self.mkExpr(CVC4.EQUAL, d, zero)
            return self.mkExpr(CVC4.ITE, test, n, base)

    def walk_bv_lshl(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_SHL, args[0], args[1])

    def walk_bv_lshr(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_LSHR, args[0], args[1])

    def walk_bv_rol(self, formula, args, **kwargs):
        ext = self.mkConst(CVC4.BitVectorRotateLeft(formula.bv_rotation_step()))
        return self.mkExpr(CVC4.BITVECTOR_ROTATE_LEFT, ext, args[0])

    def walk_bv_ror(self, formula, args, **kwargs):
        ext = self.mkConst(CVC4.BitVectorRotateRight(formula.bv_rotation_step()))
        return self.mkExpr(CVC4.BITVECTOR_ROTATE_RIGHT, ext, args[0])

    def walk_bv_zext(self, formula, args, **kwargs):
        ext = self.mkConst(CVC4.BitVectorZeroExtend(formula.bv_extend_step()))
        return self.mkExpr(CVC4.BITVECTOR_ZERO_EXTEND, ext, args[0])

    def walk_bv_sext(self, formula, args, **kwargs):
        ext = self.mkConst(CVC4.BitVectorSignExtend(formula.bv_extend_step()))
        return self.mkExpr(CVC4.BITVECTOR_SIGN_EXTEND, ext, args[0])

    def walk_bv_slt(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_SLT, args[0], args[1])

    def walk_bv_sle(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_SLE, args[0], args[1])

    def walk_bv_comp(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_COMP, args[0], args[1])

    def walk_bv_sdiv(self, formula, args, **kwargs):
        # Force deterministic semantics of division by 0
        # If the denominator is bv0, then the result is:
        #   * ~0 (if the numerator is signed >= 0)
        #   * 1 (if the numerator is signed < 0)
        n, d = args
        # sign_expr : ( 0 s<= n ) ? ~0 : 1 )
        zero = self.mkConst(CVC4.BitVector(formula.bv_width(), CVC4.Integer("0")))
        notzero = self.mkExpr(CVC4.BITVECTOR_NOT, zero)
        one = self.mkConst(CVC4.BitVector(formula.bv_width(), CVC4.Integer("1")))
        is_gt_zero = self.mkExpr(CVC4.BITVECTOR_SLE, zero, n)
        sign_expr = self.mkExpr(CVC4.ITE, is_gt_zero, notzero, one)
        base = self.mkExpr(CVC4.BITVECTOR_SDIV, n, d)
        if d.isConst():
            v = d.getConstBitVector().getValue().toString()
            if v == "0":
                return sign_expr
            else:
                return base
        else:
            # (d == 0) ? sign_expr : base
            is_zero = self.mkExpr(CVC4.EQUAL, d, zero)
            return self.mkExpr(CVC4.ITE, is_zero, sign_expr, base)

    def walk_bv_srem(self, formula, args, **kwargs):
        # Force deterministic semantics of reminder by 0
        # If the denominator is bv0, then the result is the numerator
        n, d = args
        if d.isConst():
            v = d.getConstBitVector().getValue().toString()
            if v == "0":
                return n
            else:
                return self.mkExpr(CVC4.BITVECTOR_SREM, n, d)
        else:
            # (d == 0) ? n : n bvurem d
            base = self.mkExpr(CVC4.BITVECTOR_SREM, n, d)
            zero = self.mkConst(CVC4.BitVector(formula.bv_width(), CVC4.Integer("0")))
            test = self.mkExpr(CVC4.EQUAL, d, zero)
            return self.mkExpr(CVC4.ITE, test, n, base)

    def walk_bv_ashr(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.BITVECTOR_ASHR, args[0], args[1])

    def walk_str_constant(self, formula, args, **kwargs):
        return self.mkConst(CVC4.CVC4String(formula.constant_value()))

    def walk_str_length(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_LENGTH, args[0])

    def walk_str_concat(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_CONCAT, args)

    def walk_str_contains(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_STRCTN, args[0], args[1])

    def walk_str_indexof(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_STRIDOF, args[0], args[1], args[2])

    def walk_str_replace(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_STRREPL, args[0], args[1], args[2])

    def walk_str_substr(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_SUBSTR, args[0], args[1], args[2])

    def walk_str_prefixof(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_PREFIX, args[0], args[1])

    def walk_str_suffixof(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_SUFFIX, args[0], args[1])

    def walk_str_to_int(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_STOI, args[0])

    def walk_int_to_str(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_ITOS, args[0])

    def walk_str_charat(self, formula, args, **kwargs):
        return self.mkExpr(CVC4.STRING_CHARAT, args[0], args[1])

    def _type_to_cvc4(self, tp):
        if tp.is_bool_type():
            return self.boolType
        elif tp.is_real_type():
            return self.realType
        elif tp.is_int_type():
            return self.intType
        elif tp.is_function_type():
            stps = [self._type_to_cvc4(x) for x in tp.param_types]
            rtp = self._type_to_cvc4(tp.return_type)
            return self.cvc4_exprMgr.mkFunctionType(stps, rtp)
        elif tp.is_array_type():
            # Recursively convert the types of index and elem
            idx_cvc_type = self._type_to_cvc4(tp.index_type)
            elem_cvc_type = self._type_to_cvc4(tp.elem_type)
            return self.cvc4_exprMgr.mkArrayType(idx_cvc_type, elem_cvc_type)
        elif tp.is_bv_type():
            return self.cvc4_exprMgr.mkBitVectorType(tp.width)
        elif tp.is_string_type():
            return self.stringType
        elif tp.is_custom_type():
            return self.cvc4_exprMgr.mkSort(str(tp))
        else:
            raise NotImplementedError("Unsupported type: %s" % tp)

    def _cvc4_type_to_type(self, type_):
        if type_.isBoolean():
            return types.BOOL
        elif type_.isInteger():
            return types.INT
        elif type_.isReal():
            return types.REAL
        elif type_.isArray():
            # Casting Type into ArrayType
            type_ = CVC4.ArrayType(type_)
            # Recursively convert the types of index and elem
            idx_type = self._cvc4_type_to_type(type_.getIndexType())
            elem_type = self._cvc4_type_to_type(type_.getConstituentType())
            return types.ArrayType(idx_type, elem_type)
        elif type_.isBitVector():
            # Casting Type into BitVectorType
            type_ = CVC4.BitVectorType(type_)
            return types.BVType(type_.getSize())
        elif type_.isFunction():
            # Casting Type into FunctionType
            type_ = CVC4.FunctionType(type_)
            return_type = type_.getRangeType()
            param_types = tuple(
                self._cvc4_type_to_type(ty) for ty in type_.getArgTypes()
            )
            return types.FunctionType(return_type, param_types)
        else:
            raise NotImplementedError("Unsupported type: %s" % type_)

    def _rename_bound_variables(self, formula, variables):
        """Bounds the variables in formula.

        Returns a tuple (new_formula, new_var_list) in which the old
        variables have been replaced by the new variables in the list.
        """
        mkBoundVar = self.cvc4_exprMgr.mkBoundVar
        new_var_list = [
            mkBoundVar(x.symbol_name(), self._type_to_cvc4(x.symbol_type()))
            for x in variables
        ]
        old_var_list = [self.walk_symbol(x, []) for x in variables]
        new_formula = formula.substitute(old_var_list, new_var_list)
        return (new_formula, new_var_list)
